#!/usr/bin/python

from SmartMeshSDK.ApiDefinition import ApiDefinition
from SmartMeshSDK.ApiDefinition.ByteArraySerializer import ByteArraySerializer

class IpMoteDefinition(ApiDefinition.ApiDefinition):
    '''
    \ingroup ApiDefinition
    
    \brief  API definition for the IP mote.
   
    \note   This class inherits from ApiDefinition. It redefines the attributes of
            its parents class, but inherits the methods.
    '''
    
    STRING    = ApiDefinition.FieldFormats.STRING
    BOOL      = ApiDefinition.FieldFormats.BOOL
    INT       = ApiDefinition.FieldFormats.INT
    INTS      = ApiDefinition.FieldFormats.INTS
    HEXDATA   = ApiDefinition.FieldFormats.HEXDATA
    ARRAY     = ApiDefinition.FieldFormats.ARRAY
    RC        = ApiDefinition.ApiDefinition.RC
    SUBID1    = ApiDefinition.ApiDefinition.SUBID1
    SUBID2    = ApiDefinition.ApiDefinition.SUBID2
    RC_OK     = ApiDefinition.ApiDefinition.RC_OK
    
    def __init__(self, array2scalar = True):
        ApiDefinition.ApiDefinition.__init__(self, array2scalar)
        self.serializer = ByteArraySerializer.ByteArraySerializer(self)
    
    def default_serializer(self,commandArray,fieldsToFill):
        '''
        \brief IpMgrDefinition-specific implementation of default serializer
        '''
        return self.serializer.serialize(commandArray,fieldsToFill)
        
    def deserialize(self,type,cmdId,byteArray):
        '''
        \brief IpMgrDefinition-specific implementation of deserializer
        '''
        return self.serializer.deserialize(type,cmdId,byteArray)
    
    # We redefine this attribute inherited from ApiDefinition. See
    # ApiDefinition for a full description of the structure of this field.
    fieldOptions = {
        RC: [
            [0x00, 'RC_OK',                    'Command completed successfully'],
            [0x01, 'RC_ERROR',                 'Processing error'],
            [0x03, 'RC_BUSY',                  'Device currently unavailable to perform the operation'],
            [0x04, 'RC_INVALID_LEN',           'Invalid length'],
            [0x05, 'RC_INVALID_STATE',         'Invalid state'],
            [0x06, 'RC_UNSUPPORTED',           'Unsupported command or operation'],
            [0x07, 'RC_UNKNOWN_PARAM',         'Unknown parameter'],
            [0x08, 'RC_UNKNOWN_CMD',           'Unknown command'],
            [0x09, 'RC_WRITE_FAIL',            'Couldn\'t write to persistent storage'],
            [0x0A, 'RC_READ_FAIL',             'Couldn\'t read from persistent storage'],
            [0x0B, 'RC_LOW_VOLTAGE',           'Low voltage detected'],
            [0x0C, 'RC_NO_RESOURCES',          'Couldn\'t process command due to low resources (e.g. no buffers)'],
            [0x0D, 'RC_INCOMPLETE_JOIN_INFO',  'Incomplete configuration to start joining'],
            [0x0E, 'RC_NOT_FOUND',             'Resource not found'],
            [0x0F, 'RC_INVALID_VALUE',         'Invalid value supplied'],
            [0x10, 'RC_ACCESS_DENIED',         'Access to resource or command is denied'],
            [0x12, 'RC_ERASE_FAIL',            'Erase operation failed'],
        ],
        'serviceType': [
            [0,    'bandwidth',             'Bandwidth-type service'],
        ],
        'serviceState': [
            [0,    'completed',             'Service request completed (idle)'],
            [1,    'pending',               'Service request pending'],
        ],
        'protocolType': [
            [0,    'udp',                   'User Datagram Protocol (UDP)'],
        ],
        'packetPriority': [
            [0,    'low',                   'Low priority'],
            [1,    'medium',                'Medium priority'],
            [2,    'high',                  'High priority'],
        ],
        'moteState': [
            [0,    'init',                  'Initializing'],
            [1,    'idle',                  'Idle, ready to be configured or join'],
            [2,    'searching',             'Searching for network'],
            [3,    'negotiating',           'Sent join request'],
            [4,    'connected',             'Received at least one packet from the Manager'],
            [5,    'operational',           'Configured by Manager and ready to send data'],
            [6,    'disconnected',          'Disconnected from the network'],
            [7,    'radiotest',             'The mote is in radio test mode'],
            [8,    'promiscuous listen',    'The mote received search command and is in promiscuous listen mode'],
            [9,    'blink',                 'The mote received a Blink command and will stay for 60s without any MAC-layer acks'],
        ],
        'moteEvents': [
            [0x0001, 'boot',                'The mote booted up'],
            [0x0002, 'alarmChange',         'Alarm(s) were opened or closed'],
            [0x0004, 'timeChange',          'UTC time-mapping on the mote changed'],
            [0x0008, 'joinFail',            'Join operation failed'],
            [0x0010, 'disconnected',        'The mote disconnected from the network'],
            [0x0020, 'operational',         'Mote has connection to the network and is ready to send data'],
            [0x0080, 'svcChange',           'Service allocation has changed'],
            [0x0100, 'joinStarted',          'Mote started joining the network'],
        ],
        'moteAlarms': [
            [0x01, 'nvError',               'Detected an error in persistent configuration storage (NV)'],
            [0x04, 'otpError',              'Detected an error in calibration or bsp data in flash'],
        ],
        'radioTestTypes' : [
            [0x00, 'packet',                'Packet transmission'],
            [0x01, 'cm',                    'Continuous modulation'],
            [0x02, 'cw',                    'Continuous wave'],
            [0x03, 'pkcca',                 'Packet test with clear channel assesment'],
        ],
        'packetTransmitStatus': [
            [0x00, 'ok',                    'Packet sent into the network'],
            [0x01, 'fail',                  'Packet dropped'],
        ],
        'channel': [
            [0,    '2.405GHz',              ''],
            [1,    '2.410GHz',              ''],
            [2,    '2.415GHz',              ''],
            [3,    '2.420GHz',              ''],
            [4,    '2.425GHz',              ''],
            [5,    '2.430GHz',              ''],
            [6,    '2.435GHz',              ''],
            [7,    '2.440GHz',              ''],
            [8,    '2.445GHz',              ''],
            [9,    '2.450GHz',              ''],
            [10,   '2.455GHz',              ''],
            [11,   '2.460GHz',              ''],
            [12,   '2.465GHz',              ''],
            [13,   '2.470GHz',              ''],
            [14,   '2.475GHz',              ''],
            [15,   '2.480GHz',              ''],
        ],
        'ccaMode': [
            [0,    'off',                   ''],
            [1,    'on',                    ''],
        ],
        'mobilityType': [
            [0x00, 'UNUSED',                ''],
            [0x01, 'KNOWN',                 ''],
            [0x02, 'UNKNOWN',               ''],
            [0x03, 'MOBILE',                ''],
        ],
        'seqSize': [
            [0,    '0',                     ''],
            [1,    '1',                     ''],
            [2,    '2',                     ''],
            [3,    '3',                     ''],
            [4,    '4',                     ''],
            [5,    '5',                     ''],
            [6,    '6',                     ''],
            [7,    '7',                     ''],
            [8,    '8',                     ''],
            [9,    '9',                     ''],
            [10,   '10',                    ''],
        ],
    }
    
    subCommandsSet = [
        {
            'id'         : 0x01,
            'name'       : 'macAddress',
            'description': 'This command allows user to overwrite the manufacturer-assigned MAC address of the mote. The new value takes effect after the mote resets.',
            'request'    : [
                ['macAddress',              HEXDATA,  8,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_WRITE_FAIL'              : 'Couldn\'t update persistent storage',
            },
        },
        {
            'id'         : 0x02,
            'name'       : 'joinKey',
            'description': "The setParameter<joinKey> command may be used to set the join key in mote's persistent storage. Join keys are used by motes to establish secure connection with the network. The join key is used at next join.\n\nReading the joinKey parameter is prohibited for security reasons.",
            'request'    : [
                ['joinKey',                 HEXDATA,  16,  None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_WRITE_FAIL'              : 'Could not write the key to storage',
            },
        },
        {
            'id'         : 0x03,
            'name'       : 'networkId',
            'description': "This command may be used to set the Network ID of the mote. This setting is persistent and is used on next join attempt.\n\nAs of version 1.4.x, a network ID of 0xFFFF can be used to indicate that the mote should join the first network heard.\n\n0xFFFF is never used over the air as a valid network ID - you should not set the Manager's network ID to 0xFFFF.",
            'request'    : [
                ['networkId',               INT,      2,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_WRITE_FAIL'              : 'Could not store the parameter',
            },
        },
        {
            'id'         : 0x04,
            'name'       : 'txPower',
            'description': 'This command sets the radio output power. This setting is persistent. The command may be issued at any time and takes effect on next transmission. Refer to product datasheets for supported RF output power values. If the provided txPower does not match an appropriate value for the hardware, the radio driver will select the nearest appropriate value. The nearest appropriate value varies depending on the hardware and calibration. The getParameter<txPower> command will return the value selected by the radio driver.\n\n\n\nFor example, if the part has a typical RF output power of +8 dBm when the power amplifier (PA) is enabled, then set the txPower parameter to 8 to enable the PA. Similarly, if the part has a typical RF output power of 0 dBm when the PA is disabled, then set the txPower parameter to 0 to turn off the PA. Similarly, calling this function with a value of 10 on a part, such as the LTC-5800, that supports a maximum of +8 dBm will result in a setting of +8.',
            'request'    : [
                ['txPower',                 INTS,     1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed sucessfuly',
               'RC_INVALID_VALUE'           : 'The requested power is not supported',
            },
        },
        {
            'id'         : 0x06,
            'name'       : 'joinDutyCycle',
            'description': 'The setParameter<joinDutyCycle> command allows the microprocessor to control the ratio of active listen time to doze time (a low-power radio state) during the period when the mote is searching for the network. If you desire a faster join time at the risk of higher power consumption, use the setParameter<joinDutyCycle> command to increase the join duty cycle up to 100%. This setting is persistent and takes effect immediately if the device is searching for network.',
            'request'    : [
                ['dutyCycle',               INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x0B,
            'name'       : 'eventMask',
            'description': 'The setParameter<eventMask> command allows the microprocessor to selectively subscribe to event notifications. The default value of eventMask at mote reset is all 1s - all events are enabled. This setting is not persistent.\n\nNew event types may be added in future revisions of mote software. It is recommended that the client code only subscribe to known events and gracefully ignore all unknown events.',
            'request'    : [
                ['eventMask',               HEXDATA,  4,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x15,
            'name'       : 'OTAPLockout',
            'description': 'This command allows the microprocessor to control whether Over-The-Air Programming (OTAP) of motes is allowed. This setting is persistent and takes effect immediately.',
            'request'    : [
                ['mode',                    BOOL,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_INVALID_VALUE'           : 'Invalid value of mode parameter',
               'RC_WRITE_FAIL'              : 'Couldn\'t update persistent storage',
            },
        },
        {
            'id'         : 0x1D,
            'name'       : 'routingMode',
            'description': 'This command allows the microprocessor to control whether the mote will become a router once joined the network. If disabled, the manager will keep the mote a leaf node.',
            'request'    : [
                ['mode',                    BOOL,     1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_INVALID_VALUE'           : 'Invalid value of mode parameter',
            },
        },
        {
            'id'         : 0x1F,
            'name'       : 'powerSrcInfo',
            'description': 'This command allows the microprocessor to configure power source information on the device. This setting is persistent and is used at network join time.',
            'request'    : [
                ['maxStCurrent',            INT,      2,   None],
                ['minLifetime',             INT,      1,   None],
                ['currentLimit_0',          INT,      2,   None],
                ['dischargePeriod_0',       INT,      2,   None],
                ['rechargePeriod_0',        INT,      2,   None],
                ['currentLimit_1',          INT,      2,   None],
                ['dischargePeriod_1',       INT,      2,   None],
                ['rechargePeriod_1',        INT,      2,   None],
                ['currentLimit_2',          INT,      2,   None],
                ['dischargePeriod_2',       INT,      2,   None],
                ['rechargePeriod_2',        INT,      2,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x22,
            'name'       : 'advKey',
            'description': 'Sets the Advertisement MIC key - this key is used to authenticate advertisements, and can be set per vendor/installation to prevent unauthorized devices from being able to respond to advertisements. If changed, it must match that set on the corresponding AP (using mset on the manager CLI) in order for the mote to join. It can be reset to default via the clearNV command.',
            'request'    : [
                ['advKey',                  HEXDATA,  16,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_WRITE_FAIL'              : 'Couldn\'t update persistent storage',
            },
        },
        {
            'id'         : 0x24,
            'name'       : 'autoJoin',
            'description': 'This command allows the microprocessor to change between automatic and manual joining by the mote\'s networking stack. In manual mode, an explicit join command from the application is required to initiate joining. This setting is persistent and takes effect after mote reset.\n\nNote that auto join mode must not be set if the application is also configured to join (e.g combining \'auto join\' with \'master\' mode will result in mote not joining).',
            'request'    : [
                ['mode',                    BOOL,     1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_WRITE_FAIL'              : 'Couldn\'t update persistent storage',
               'RC_INVALID_VALUE'           : 'Invalid value specified for mode',
            },
        },
        {
            'id'         : 0x29,
            'name'       : 'antGain',
            'description': 'The setParameter<antGain> command sets the antenna gain of the system (to properly calculate radiated power). Defaults to 2 dBi if not set.\n\nNote: This parameter is available in devices running mote software >=1.4.x',
            'request'    : [
                ['antGain',                 INTS,     1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_WRITE_FAIL'              : 'Couldn\'t update persistent storage',
            },
        },
        {
            'id'         : 0x2a,
            'name'       : 'euCompliantMode',
            'description': 'The setParameter<euCompliantMode> command may be used to set the EN 300 328 compliance mode that is used by the device. When enabled, the mote may skip some transmit opportunities to remain within average power limits. Motes below +10 dBm radiated power do not need to duty cycle to meet EN 300 328 requirements.\n\nNote: This parameter is available in devices running mote software >=1.4.x',
            'request'    : [
                ['euCompliantMode',         INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_WRITE_FAIL'              : 'Couldn\'t update persistent storage',
            },
        },
    ]
    
    subCommandsGet = [
        {
            'id'         : 0x01,
            'name'       : 'macAddress',
            'description': "This command returns the user defined MAC address of the device. By default, the MAC address returned is 00's if not overwritten using the setParameter<macAddress> command. In mote versions prior to 1.4.1, it returned the factory MAC address.",
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['macAddress',          HEXDATA,  8,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x03,
            'name'       : 'networkId',
            'description': 'This command returns the network id stored in mote\'s persistent storage.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['networkId',           INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x04,
            'name'       : 'txPower',
            'description': 'This command gets the radio output power in dBm, excluding any antenna gain. This value corresponds to the actual output power used by the radio driver and may not be the same as the input value entered with the setParameter<txPower>, which will set to nearest if the value entered is not supported by the hardware.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['txPower',             INTS,     1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x06,
            'name'       : 'joinDutyCycle',
            'description': 'This command allows user to retrieve current value of joinDutyCycle parameter.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['joinDutyCycle',       INT,      1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x0B,
            'name'       : 'eventMask',
            'description': 'getParameter<eventMask> allows the microprocessor to read the currently subscribed-to event types.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['eventMask',           HEXDATA,  4,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x0C,
            'name'       : 'moteInfo',
            'description': 'The getParameter<moteInfo> command returns static information about the motes hardware and network stack software.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['apiVersion',          INT,      1,   None],
                    ['serialNumber',        HEXDATA,  8,   None],
                    ['hwModel',             INT,      1,   None],
                    ['hwRev',               INT,      1,   None],
                    ['swVerMajor',          INT,      1,   None],
                    ['swVerMinor',          INT,      1,   None],
                    ['swVerPatch',          INT,      1,   None],
                    ['swVerBuild',          INT,      2,   None],
                    ['bootSwVer',           INT,      1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x0D,
            'name'       : 'netInfo',
            'description': 'The getParameter<networkInfo> command may be used to retrieve the mote\'s network-related parameters.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['macAddress',          HEXDATA,  8,   None],
                    ['moteId',              INT,      2,   None],
                    ['networkId',           INT,      2,   None],
                    ['slotSize',            INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x0E,
            'name'       : 'moteStatus',
            'description': 'The getParameter<moteStatus> command is used to retrieve current mote state and other dynamic information.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['state',               INT,      1,   'moteState'],
                    ['reserved_0',          INT,      1,   None],
                    ['reserved_1',          INT,      2,   None],
                    ['numParents',          INT,      1,   None],
                    ['alarms',              INT,      4,   None],
                    ['reserved_2',          INT,      1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x0F,
            'name'       : 'time',
            'description': 'The getParameter<time> command may be used to request the current time on the mote. The mote reports time at the moment it is processing the command, so the information includes variable delay. For more precise time information consider using TIMEn pin (see timeIndication).',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['upTime',              INT,      4,   None],
                    ['utcSecs',             INT,      8,   None],
                    ['utcUsecs',            INT,      4,   None],
                    ['asn',                 HEXDATA,  5,   None],
                    ['asnOffset',           INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x10,
            'name'       : 'charge',
            'description': 'The getParameter<charge> command retrieves the charge consumption of the mote since the last reset.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['qTotal',              INT,      4,   None],
                    ['upTime',              INT,      4,   None],
                    ['tempInt',             INTS,     1,   None],
                    ['tempFrac',            INT,      1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x11,
            'name'       : 'testRadioRxStats',
            'description': 'The getParameter<testRadioRxStats> command retrieves statistics for the latest radio test performed using the testRadioRx command. The statistics show the number of good and bad packets (CRC failures) received during the test',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['rxOk',                INT,      2,   None],
                    ['rxFailed',            INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x15,
            'name'       : 'OTAPLockout',
            'description': 'This command reads the current state of OTAP lockout, i.e. whether over-the-air upgrades of software are permitted on this mote.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['mode',                BOOL,     1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x17,
            'name'       : 'moteId',
            'description': 'This command retrieves the mote\'s Mote ID. If the mote is not in the network, value of 0 is returned.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['moteId',              INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x18,
            'name'       : 'ipv6Address',
            'description': 'This command allows the microprocessor to read IPV6 address assigned to the mote. Before the mote has an assigned address it will return all 0s.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['ipv6Address',         HEXDATA,  16,  None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x1D,
            'name'       : 'routingMode',
            'description': 'This command allows the microprocessor to retrieve the current routing mode of the mote.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['routingMode',         BOOL,     1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x1E,
            'name'       : 'appInfo',
            'description': 'Get the application (as opposed to the network stack) version information.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['vendorId',            INT,      2,   None],
                    ['appId',               INT,      1,   None],
                    ['appVer',              HEXDATA,  5,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x1F,
            'name'       : 'powerSrcInfo',
            'description': 'This command allows the microprocessor to read a mote\'s power source settings.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['maxStCurrent',        INT,      2,   None],
                    ['minLifetime',         INT,      1,   None],
                    ['currentLimit_0',      INT,      2,   None],
                    ['dischargePeriod_0',   INT,      2,   None],
                    ['rechargePeriod_0',    INT,      2,   None],
                    ['currentLimit_1',      INT,      2,   None],
                    ['dischargePeriod_1',   INT,      2,   None],
                    ['rechargePeriod_1',    INT,      2,   None],
                    ['currentLimit_2',      INT,      2,   None],
                    ['dischargePeriod_2',   INT,      2,   None],
                    ['rechargePeriod_2',    INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x24,
            'name'       : 'autoJoin',
            'description': 'This command allows the microprocessor to retrieve the current autoJoin setting.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['autoJoin',            BOOL,     1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'The command completed successfully',
            },
        },
        {
            'id'         : 0x29,
            'name'       : 'antGain',
            'description': 'The getParameter<antGain> command retrieves the antenna gain used by the system (to properly calculate radiated power). Defaults to 2 dBi.\n\nNote: This parameter is available in devices running mote software >=1.4.x',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['antGain',             INTS,     1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x2a,
            'name'       : 'euCompliantMode',
            'description': 'The getParameter<euCompliantMode> command may be used to retrieve the EN 300 328 compliance mode that is used by the device. When enabled, the mote may skip some transmit opportunities to remain within average power limits. Motes below +10 dBm radiated power do not need to duty cycle to meet EN 300 328 requirements.\n\nNote: This parameter is available in devices running mote software >=1.4.x',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['compMode',            INT,      1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
            },
        },
        {
            'id'         : 0x2c,
            'name'       : 'entropy',
            'description': 'The getParameter<entropy> command may be used to retrieve a 16-byte block of random data. The data is obtained from thermal noise in the LTC5800 receive signal chain with the radio front-end disabled - as such, it can only be called when the mote is in the idle state. Thus while it is suitable for cryptographic operations, it is recommended to be used as a seed for a DRBG because of this limitation.\n\nThis parameter is available in devices running mote software 1.4.0 or later.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    ['entropy',         HEXDATA,      16,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_INVALID_STATE'           : 'The mote is in invalid state to retrieve entropy',
            },
        },
    ]
    
    # We redefine this attribute inherited from ApiDefinition. See
    # ApiDefinition for a full description of the structure of this field.
    commands = [
        {
            'id'         : 0x01,
            'name'       : 'setParameter',
            'description': 'Set Parameter',
            'request'    : [
                [SUBID1,                    INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    [SUBID1,                INT,      1,   None],
                ]
            },
            'subCommands': subCommandsSet,
        },
        {
            'id'         : 0x02,
            'name'       : 'getParameter',
            'description': 'Get Parameter',
            'request'    : [
                [SUBID1,                    INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    [SUBID1,                INT,      1,   None],
                ]
            },
            'subCommands': subCommandsGet,
        },
        {
            'id'         : 0x06,
            'name'       : 'join',
            'description': 'The join command requests that mote start searching for the network and attempt to join.The mote must be in the Idle state or the Promiscuous Listen state (see search) for this command to be valid. Note that the join time will be affected by the maximum current setting.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
               'RC_INVALID_STATE'           : 'The mote is in invalid state to start join operation',
               'RC_INCOMPLETE_JOIN_INFO'    : 'Incomplete configuration to start joining',
            },
        },
        {
            'id'         : 0x07,
            'name'       : 'disconnect',
            'description': 'The disconnect command requests that the mote initiate disconnection from the network. After disconnection completes, the mote will generate a disconnected event, and proceed to reset. If the mote is not in the network, the disconnected event will be generated immediately. This command will return an error if the mote is not presently joining or joined to a network.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
            },
        },
        {
            'id'         : 0x08,
            'name'       : 'reset',
            'description': 'The reset command initiates a soft-reset of the device. The device will initiate the reset sequence shortly after sending out the response to this command. Resetting a mote directly can adversely impact its descendants; to disconnect gracefully from the network, use the disconnect command',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
            },
        },
        {
            'id'         : 0x09,
            'name'       : 'lowPowerSleep',
            'description': 'The lowPowerSleep command shuts down all peripherals and places the mote into deep sleep mode. The command executes after the mote sends its response. The mote enters deep sleep within two seconds after the command executes. The command may be issued at any time and will cause the mote to interrupt all in-progress network operation. To achieve a graceful disconnect, use the disconnect command before using the lowPowerSleep command. A hardware reset is required to bring a mote out of deep sleep mode.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
            },
        },
        {
            'id'         : 0x0C,
            'name'       : 'testRadioRx',
            'description': 'The testRadioRx command clears all previously collected statistics and initiates a test of radio reception for the specified channel and duration. During the test, the mote keeps statistics about the number of packets received (with and without error). The test results may be retrieved using the getParameter<testRadioRxStats> command. The testRadioRx command may only be issued in Idle mode. The mote must be reset (either hardware or software reset) after radio tests are complete and prior to joining.\n\nStation ID is available in IP Mote 1.4.0 or later, and WirelessHART Mote 1.1.2 or later. The station ID is a user selectable value used to isolate traffic if multiple tests are running in the same radio space. It must be set to match the station ID used by the transmitter.\n\n\n\nChannel numbering is 0-15, corresponding to IEEE 2.4 GHz channels 11-26.',
            'request'    : [
                ['channelMask',             HEXDATA,  2,   None],
                ['time',                    INT,      2,   None],
                ['stationId',               INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,  1,   True]
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
               'RC_INVALID_VALUE'           : 'The mote is not in Idle state',
               'RC_BUSY'                    : 'Another test operation in progress',
            },
        },
        {
            'id'         : 0x10,
            'name'       : 'clearNV',
            'description': 'The clearNV command resets the motes non-volatile memory (NV) to its factory-default state. See User Guide for detailed information about the default values. Since many parameters are read by the mote only at power-up, this command should be followed up by mote reset.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_WRITE_FAIL'              : 'Flash operation failed',
            },
        },
        {
            'id'         : 0x11,
            'name'       : 'requestService',
            'description': 'The requestService command may be used to request a new or changed service level to a destination device in the mesh. This command may only be used to update the service to a device with an existing connection (session).\n\nWhenever a change in bandwidth assignment occurs, the application receives a serviceChanged event that it can use as a trigger to read the new service allocation.',
            'request'    : [
                ['destAddr',                INT,      2,   None],
                ['serviceType',             INT,      1,   True],
                ['value',                   INT,      4,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Service request accepted',
            },
        },
        {
            'id'         : 0x12,
            'name'       : 'getServiceInfo',
            'description': 'The getServiceInfo command returns information about the service currently allocated to the mote.',
            'request'    : [
                ['destAddr',                INT,      2,   None],
                ['type',                    INT,      1,   'serviceType'],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['destAddr',            HEXDATA,  2,   None],
                    ['type',                INT,      1,   'serviceType'],
                    ['state',               INT,      1,   'serviceState'],
                    ['value',               INT,      4,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command successfully completed',
            },
        },
        {
            'id'         : 0x15,
            'name'       : 'openSocket',
            'description': 'The openSocket command creates an endpoint for IP communication and returns an ID for the socket.',
            'request'    : [
                ['protocol',                INT,      1,   'protocolType'],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['socketId',            INT,      1,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfuly',
               'RC_NO_RESOURCES'            : 'Couldn\'t create new socket due to resource availability',
            },
        },
        {
            'id'         : 0x16,
            'name'       : 'closeSocket',
            'description': 'Close the previously open socket.',
            'request'    : [
                ['socketId',                INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Socket successfully closed',
               'RC_NOT_FOUND'               : 'Socket ID not found',
            },
        },
        {
            'id'         : 0x17,
            'name'       : 'bindSocket',
            'description': 'Bind a previously opened socket to a port. When a socket is created, it is only given a protocol family, but not assigned a port. This association must be performed before the socket can accept connections from other hosts.',
            'request'    : [
                ['socketId',                INT,      1,   None],
                ['port',                    INT,      2,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Operation completed successfully',
               'RC_NOT_FOUND'               : 'Invalid socket ID',
            },
        },
        {
            'id'         : 0x18,
            'name'       : 'sendTo',
            'description': 'Send a packet into the network. If the command returns RC_OK, the mote has accepted the packet and has queued it up for transmission. A txDone notification will be issued when the packet has been sent, if and only if the packet ID passed in this command is different from 0xffff. You can set the packet ID to any value. The notification will contain the packet ID of the packet just sent, allowing association of the notification with a particular packet. The destination port should be in the range 0xF0B8-F0BF (61624-61631) to maximize payload.',
            'request'    : [
                ['socketId',                INT,      1,   None],
                ['destIP',                  HEXDATA,  16,  None],
                ['destPort',                INT,      2,   None],
                ['serviceType',             INT,      1,   True],
                ['priority',                INT,      1,   'packetPriority'],
                ['packetId',                INT,      2,   None],
                ['payload',                 HEXDATA,  None,None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Packet was queued up for transmission',
               'RC_NO_RESOURCES'            : 'No queue space to accept the packet',
            },
        },
        {
            'id'         : 0x24,
            'name'       : 'search',
            'description': 'The search command requests that mote start listening for advertisements and report those heard from any network without attempting to join. This is called the Promiscuous Listen state. The mote must be in the Idle state for this command to be valid. The search state can be exited by issuing the join command or the reset command.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
               'RC_INVALID_STATE'           : 'The mote is in invalid state to start search operation',
            },
        },
        {
            'id'         : 0x28,
            'name'       : 'testRadioTxExt',
            'description': 'The testRadioTxExt command allows the microprocessor to initiate a radio transmission test. This command may only be issued prior to the mote joining the network. Three types of transmission tests are supported:\n\n- Packet transmission\n- Continuous modulation\n- Continuous wave (unmodulated signal)\n\nIn a packet transmission test, the mote generates a repeatCnt number of packet sequences. Each sequence consists of up to 10 packets with configurable size and delays. Each packet starts with a PHY preamble (5 bytes), followed by a PHY length field (1 byte), followed by data payload of up to 125 bytes, and finally a 2-byte 802.15.4 CRC at the end. Byte 0 of the payload contains stationId of the sender. Bytes 1 and 2 contain the packet number (in big-endian format) that increments with every packet transmitted. Bytes 3..N contain a counter (from 0..N-3) that increments with every byte inside payload. Transmissions occur on the set of channels defined by chanMask , selected inpseudo-randomorder.\n\nIn a continuous modulation test, the mote generates continuous pseudo-random modulated signal, centered at the specified channel. The test is stopped by resetting the mote.\n\nIn a continuous wave test, the mote generates an unmodulated tone, centered at the specified channel. The test tone is stopped by resetting the mote.\n\nThe testRadioTxExt command may only be issued when the mote is in Idle mode, prior to its joining the network. The mote must be reset (either hardware or software reset) after radio tests are complete and prior to joining.\n\nThe station ID is a user selectable value. It is used in packet tests so that a receiver can identify packets from this device in cases where there may be multiple tests running in the same radio space. This field is not used for CM or CW tests. See testRadioRX (SmartMesh IP) or testRadioRxExt (SmartMesh WirelessHART).\n\n\n\nChannel numbering is 0-15, corresponding to IEEE 2.4 GHz channels 11-26.',
            'request'    : [
                ['testType',                INT,      1,   'radioTestTypes'],
                ['chanMask',                HEXDATA,  2,   None],
                ['repeatCnt',               INT,      2,   None],
                ['txPower',                 INTS,     1,   None],
                ['seqSize',                 INT,      1,   'seqSize'],
                ['pkLen_1',                 INT,      1,   None],
                ['delay_1',                 INT,      2,   None],
                ['pkLen_2',                 INT,      1,   None],
                ['delay_2',                 INT,      2,   None],
                ['pkLen_3',                 INT,      1,   None],
                ['delay_3',                 INT,      2,   None],
                ['pkLen_4',                 INT,      1,   None],
                ['delay_4',                 INT,      2,   None],
                ['pkLen_5',                 INT,      1,   None],
                ['delay_5',                 INT,      2,   None],
                ['pkLen_6',                 INT,      1,   None],
                ['delay_6',                 INT,      2,   None],
                ['pkLen_7',                 INT,      1,   None],
                ['delay_7',                 INT,      2,   None],
                ['pkLen_8',                 INT,      1,   None],
                ['delay_8',                 INT,      2,   None],
                ['pkLen_9',                 INT,      1,   None],
                ['delay_9',                 INT,      2,   None],
                ['pkLen_10',                INT,      1,   None],
                ['delay_10',                INT,      2,   None],
                ['stationId',               INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
               'RC_INVALID_STATE'           : 'Mote is not in correct state to accept the command',
               'RC_BUSY'                    : 'Mote is executing another radio test operation',
            },
            'serializer' : 'serialize_testRadioTxExt',
        },
        {
            'id'         : 0x29,
            'name'       : 'zeroize',
            'description': 'Zeroize (zeroise) command erases flash area that is used to store configuration parameters, such as join keys. This command is intended to satisfy zeroization requirement of FIPS-140 standard. After the command executes, the mote should be reset. Available in mote >= 1.4.x\n\nThe zeroize command will render the mote inoperable. It must be re-programmed via SPI or JTAG in order to be useable.',
            'request'    : [
                ['password',                INT,      4,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command completed successfully',
               'RC_ERASE_FAIL'              : 'Flash could not be erased due to unexpected internal error',
            },
        },
        {
            'id'         : 0x2B,
            'name'       : 'socketInfo',
            'description': 'Retrieve information about a socket. Available in IP Mote 1.4.0 or later.',
            'request'    : [
                ['index',                   INT,      1,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                    ['index',               INT,      1,   None],
                    ['socketId',            INT,      1,   None],
                    ['protocol',            INT,      1,   None],
                    ['bindState',           INT,      1,   None],
                    ['port',                INT,      2,   None],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Operation completed normally',
               'RC_NOT_FOUND'               : 'A socket could not be found for the index provided',
            },
        },
        {
            'id'         : 0x2E,
            'name'       : 'blink',
            'description': 'Send a packet into the network without joining. The mote searches for a network and sends the packet. Optionally, the list of neighbors discovered during the search process is also sent, up to a maximum of four neighbors. If the discovered neighbor list is not included, the payload maximum size is 73B, and if the discovered neighbors are included, the maximum size is 58B.\n\nUpon receiving a blink command, the mote will transition to the blink state and start searching for advertisements. When it hears an advertisement, it synchronizes and continues listening briefly in efforts to discover more neighbors. After this short timeout, the mote immediately sends the data packet to one of the discovered neighbors. If the blink command is called repeatedly to send consecutive packets, the mote does not search again unless the discovered neighbor list is requested.\n\nWhen the mote successfully sends the packet, a txDone notification will be sent with status set to 0. If the mote cannot send the packet, e.g. if 60 seconds elapse without receiving a MAC-layer acknowledgement, a txDone notification is sent with status set to 1.\n\nFor Blink packets, the mote can only accept a single packet at a time. To send multiple packets, the application must wait for the txDone notification. The mote will return to low-power sleep when 60 seconds elapse without any MAC-layer acknowledgements, so to prevent the mote from sleeping, the application should send the packets much faster than this 60 second timeout. See the SmartMesh Embedded IP Manager API Guide for details on the manager-side blink notification.\n\nThe mote will stay in the blink state unless it is reset, OR a join command is issued. This was done so that the getTime command can be called at any time after at least one blink packet has been sent. Note that the clock is free-running since its last packet sent and will drift, however can be used as a relatively accurate clock source for many applications, if sending packets once a day for example.',
            'request'    : [
                ['fIncludeDscvNbrs',        INT,      1,   None],
                ['payload',                 HEXDATA,  None,None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Packet accepted',
               'RC_INVALID_STATE'           : 'Another Blink packet is in the queue to be sent',
               'RC_INVALID_LEN'             : 'The payload is too long',
            },
        },
        {
            'id'         : 0x2f,
            'name'       : 'stopSearch',
            'description': 'The stopSearch command stops the search that was started either by the join or the search command. The mote must be in either Promiscuous Listen or Search state for this command to be valid. The mote goes back to Idle state when this command is received in a valid state. Available in Mote 1.4.0 or later.',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
               'RC_INVALID_STATE'           : 'The mote is in invalid state to stop search operation',
            },
        },
        {
            'id'         : 0x30,
            'name'       : 'testRadioRxPER',
            'description': 'The testRadioRxPER command initiates Packet Error Rate (PER) test in rx mode. This command may be issued only when mote is in Idle state.\n\nThis command is available in SmartMesh IP Stack version 1.4.2.5 or later',
            'request'    : [
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
               'RC_INVALID_STATE'           : 'The mote is in invalid state to start PER test',
            },
        },
        {
            'id'         : 0x31,
            'name'       : 'testRadioTxPER',
            'description': 'The testRadioTxPER command initiates Packet Error Rate (PER) test in tx mode. This command may be issued only when mote is in Idle state.\n\nThis command is available in SmartMesh IP Stack version 1.4.2.5 or later',
            'request'    : [
                ['txPower',        INTS,      1,   None],
                ['numPackets',     INT,       2,   None],
                ['chanMask',       HEXDATA,   2,   None],
                ['numRepeat',     INT,       2,   None],
            ],
            'response'   : {
                'FIELDS':  [
                    [RC,                    INT,      1,   True],
                ],
            },
            'responseCodes': {
               'RC_OK'                      : 'Command was accepted',
               'RC_INVALID_STATE'           : 'The mote is in invalid state to start PER test',
            },
        },
    ]
    
    # We redefine this attribute inherited from ApiDefinition. See
    # ApiDefinition for a full description of the structure of this field.
    notifications = [
        {
            'id'         : 0x0D,
            'name'       : 'timeIndication',
            'description': 'The timeIndication notification applies to mote products that support a time interrupt into the mote. The time packet includes the network time and the current UTC time relative to the manager.\n\nFor LTC5800-IPM based products, driving the TIMEn pin low (assert) wakes the processor. The pin must asserted for a minimum of t strobe s. De-asserting the pin latches the time, and a timeIndication will be generated within t response ms. Refer to the LTC5800-IPM Datasheet for additional information about TIME pin usage. The processor will remain awake and drawing current while the TIMEn pin is asserted. To avoid drawing excess current, take care to minimize the duration of the TIMEn pin being asserted past t strobe minimum.',
            'response'   : {
                'FIELDS':  [
                    ['uptime',              INT,      4,   None],
                    ['utcSecs',             INT,      8,   None],
                    ['utcUsecs',            INT,      4,   None],
                    ['asn',                 HEXDATA,  5,   None],
                    ['asnOffset',           INT,      2,   None],
                    ['asnSubOffset',        INT,      2,   None],
                ],
            },
        },
        {
            'id'         : 0x0F,
            'name'       : 'events',
            'description': 'The mote sends an events notification to inform the application of new events that occurred since the previous events notification. The notification also contains up-to-date information about current mote state and any pending alarms.',
            'response'   : {
                'FIELDS':  [
                    ['events',              INT,      4,   'moteEvents'],
                    ['state',               INT,      1,   'moteState'],
                    ['alarmsList',          INT,      4,   None],
                ],
            },
        },
        {
            'id'         : 0x19,
            'name'       : 'receive',
            'description': 'Informs the application that a packet was received.',
            'response'   : {
                'FIELDS':  [
                    ['socketId',            INT,      1,   None],
                    ['srcAddr',             HEXDATA,  16,  None],
                    ['srcPort',             INT,      2,   None],
                    ['payload',             HEXDATA,  None,None],
                ],
            },
        },
        {
            'id'         : 0x24,
            'name'       : 'macRx',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['payload',             HEXDATA,  None,None],
                ],
            },
        },
        {
            'id'         : 0x25,
            'name'       : 'txDone',
            'description': 'The txDone notification informs the application that the mote has finished sending a packet. This notification will only be generated if the user has provided a valid (0x0000-0xFFFE) packet ID when calling the sendTo command.',
            'response'   : {
                'FIELDS':  [
                    ['packetId',            INT,      2,   None],
                    ['status',              INT,      1,   'packetTransmitStatus'],
                ],
            },
        },
        {
            'id'         : 0x26,
            'name'       : 'advReceived',
            'description': 'The advReceived notification is generated by the mote w hen it is in Promiscuous Listen state (see the Mote States table) and it receives an advertisement.',
            'response'   : {
                'FIELDS':  [
                    ['netId',             INT,      2,   None],
                    ['moteId',            INT,      2,   None],
                    ['rssi',              INTS,     1,   None],
                    ['joinPri',           INT,      1,   None],
                ],
            },
        },
        {
            'id'         : 0x35,
            'name'       : 'testRadioStatsPER',
            'description': 'The testRadioStatsPER notification is generated by the mote when PER test in RX mode is completed.\n\nThis command is available in SmartMesh IP Stack version 1.4.2 or later.',
            'response'   : {
                'FIELDS':  [
                    ['numRxOK',           INT,      2,   None],
                    ['numRxErr',          INT,      2,   None],
                    ['numRxInv',          INT,      2,   None],
                    ['numRxMiss',         INT,      2,   None],
                    ['perInt',            INT,      2,   None],
                    ['perFrac',           INT,      2,   None],
                ],
            },
        },
    ]
    
    def serialize_testRadioTxExt(self, commandArray, cmd_params):
        
        # start by serializing with complete API definition
        returnVal = self.default_serializer(commandArray, cmd_params)
        
        # remove the unused sequence definitions
        startIndex = 7+cmd_params['seqSize']*3
        numVals    = 10-cmd_params['seqSize']
        returnVal[1][startIndex:startIndex+numVals*3] = []
        
        return returnVal



"""
This module implements the Telemetry Health view for the application.

NOTE: This is just a dummy implementation of a MainView in order to shown the
mechanism provided by the `core_components` package.
"""
from txp.web.core_components.app_profiles import AppProfile
from txp.web.core_components.main_view import MainView
import txp.cloud.controllers.firestore as firestore_controller
import logging
from typing import Dict
import streamlit as st
import pandas as pd
from txp.web.core_components.app_component import Card

# TODO: This might be a problem. The logging configuration is being taken from `txp` package
from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class TelemetryHealthView(MainView):
    """The Telemetry Health View"""

    def __init__(self):
        super(TelemetryHealthView, self).__init__(component_key="telemetry_health_view")
        self.data = None
        self.edges = None
        self.current_machine = None
        self.current_machines_groups = None
        self.default_edge_state = "Disconnected"
        self.formatted_asset_name = None

    def get_associated_profile(self) -> AppProfile:
        return AppProfile.RESUMEN

    @classmethod
    def get_display_name(cls) -> str:
        return "Salud de Telemetría"

    def get_data(self):
        if self.data is None:
            db = firestore_controller.get_authenticated_client(
                self.app_state.authentication_info.role_service_account_credentials
            )

            data = firestore_controller.pull_recent_project_data(db, self._tenant_id)
            self.data = self._formatted_data(data)
            machines = list(self.data["machines"].keys())
            machines_groups = list(self.data["machines_groups"].keys())
            self.current_machines_groups = machines_groups[0]
            self.current_machine = machines[0]
            self.formatted_asset_name = self.formatting_readable_strings(self.current_machine)
            self.edges = self._edges_status_mapping(self.data["machines"][self.current_machine])
            db.close()

    def _build_submenu_state(self) -> Dict:
        pass

    def _build_content_state(self) -> Dict:
        pass

    def _update_machine(self):
        machine = st.session_state[f"{self.__class__.__name__}_machine_selectbox"]
        self.current_machine = machine
        self.edges = self._edges_status_mapping(self.data["machines"][machine])
        self.formatted_asset_name = self.formatting_readable_strings(self.current_machine)

    def _update_machines_groups(self):
        machines_groups = st.session_state[
            f"{self.__class__.__name__}_machine_groups_selectbox"
        ]
        self.current_machines_groups = machines_groups

    def _formatted_data(self, data):
        """
        takes the project data and returns a formatted Dict.
        """
        machines_groups = {d["name"]: d for d in data["machines_groups"]}
        machines = {d["machine_id"]: d for d in data["machines"]}
        gateways = {d["gateway_id"]: d for d in data["gateways"]}
        edges = {d["logical_id"]: d for d in data["edges"]}
        data = {
            "machines_groups": machines_groups,
            "machines": machines,
            "gateways": gateways,
            "edges": edges,
        }
        return data

    def _render_submenu(self):
        self.get_data()
        activos = list(self.data["machines"].keys())
        st.selectbox(
            label="Selecciona un activo",
            options=activos,
            key=f"{self.__class__.__name__}_machine_selectbox",
            on_change=self._update_machine,
            index=0
        )
        st.markdown("-------")
        st.markdown("**<p align='center'>Estado general</p>**", unsafe_allow_html=True)
        st.markdown(f"⚙️ **Dispositivos configurados:** {len(self.edges.Dispositivo)}")
        st.markdown(
            f"✅ **Dispositivos conectados:** {len(self.edges.loc[(self.edges['Actividad'] == 'Conectado')])+len(self.edges.loc[(self.edges['Actividad'] == 'Recolectando datos')])}"
        )
        st.markdown(
            f"📥 **Dispositivos recolectando datos:** {len(self.edges.loc[(self.edges['Actividad'] == 'Recolectando datos')])}"
        )

    def _edges_status_mapping(self, current_machine):
        """
        Given the current machine, returns a DataFrame with the edges state and gateway-id
        """
        edges = []
        edges_state = []
        gateways = []

        for e in current_machine["associated_with_edges"]:
            edge_name = e.get().to_dict().get("logical_id")
            edge_state = e.get().to_dict().get("state")
            edge_type = e.get().to_dict().get("device_type")
            if edge_type == 'ACTUATOR_DEVICE':
                pass
            else:
                if edge_state is not None:
                    driver_status = edge_state.get("driver_state")
                    edges.append(edge_name)
                    edges_state.append(self._edge_states_sp[driver_status])
                else:
                    edges.append(edge_name)
                    edges_state.append(self._edge_states_sp[self.default_edge_state])
        for g in self.data["gateways"]:
            gateway = self.data["gateways"][g]
            machines_in_gateway = gateway.get("machines")
            for machine in machines_in_gateway:
                machine_in_gateway_name = machine.get().to_dict().get("machine_id")
                if machine_in_gateway_name == current_machine.get("machine_id"):
                    gateway_name = gateway.get("gateway_id")
                    for i in range(len(edges)):
                        operation = "" + gateway_name
                        gateways.append(operation)

        if not edges_state:
            for i in range(len(edges)):
                edges_state.append(self._edge_states_sp[self.default_edge_state])
            df = pd.DataFrame(
                list(zip(edges, edges_state, gateways)),
                columns=[
                    "Dispositivo",
                    "Actividad",
                    "Conectado al gateway",
                ],
            )
            return df
        else:
            df = pd.DataFrame(
                list(zip(edges, edges_state, gateways)),
                columns=[
                    "Dispositivo",
                    "Actividad",
                    "Conectado al gateway",
                ],
            )
            return df

    def _render_content(self):
        # helper functions
        self.get_data()
        st.header("Salud de Telemetría")
        st.subheader("Resumen")
        machine = self.current_machine
        machines_groups = self.current_machines_groups

        device_state = Card(
            card_row=1,
            card_column_start=1,
            card_column_end=1,
            card_subheader="Estado de los dispositivos",
            content=self.edges.to_dict("list"),
            labels=[
                    "Dispositivo",
                    "Actividad",
                    "Conectado al gateway",
            ],
            vertical_style=False,
        )
        metadata_of_asset = Card(
            card_row=1,
            card_column_start=2,
            card_column_end=2,
            card_subheader=f"{self.formatted_asset_name}",
            content=
            f"**Grupo de activos**: {self.data['machines_groups'][machines_groups]['name']}  \n"
            f"**Manufactura**: {self.data['machines'][machine]['manufacturer']}  \n"
            f"**Modelo**: {self.data['machines'][machine]['model']}  \n",
        )
        self.render_card([metadata_of_asset, device_state])



from txp.ml.prediction_service.state_managers import state_manager as sm
import ray
from txp.common.ml.tasks import AssetState, AssetStateCondition
import logging


class MockAsset(sm.StateManager):
    states_map = {
        "Ok": AssetStateCondition.OPTIMAL,
        "Bad": AssetStateCondition.CRITICAL
    }

    def __init__(self, asset_id, tasks, events_and_states_dataset, backup_collection_name, notifications_topic,
                 credentials_str, log_level=logging.INFO):
        logging.basicConfig(level=log_level)
        super().__init__(asset_id, events_and_states_dataset, backup_collection_name, notifications_topic,
                         credentials_str)
        self.tasks = tasks

    def get_state(self, events) -> AssetState:
        event = events[0]
        ids = [event["event_id"]]
        return AssetState(events=ids,
                          asset_id=self.asset_id,
                          condition=MockAsset.states_map.get(event["event"], AssetStateCondition.UNDEFINED),
                          observation_timestamp=event["observation_timestamp"],
                          tenant_id=event["tenant_id"],
                          partition_timestamp=event["partition_timestamp"])


@ray.remote
class RayMockAsset(MockAsset):
    def __init__(self, asset_id, tasks, events_and_states_dataset, backup_collection_name, notifications_topic,
                 credentials_str, log_level=logging.INFO):
        super().__init__(asset_id, tasks, events_and_states_dataset, backup_collection_name, notifications_topic,
                         credentials_str, log_level)



#import common.iCOMOX_messages


# on_get_in_message_size(accumulated_msg)
# Parameter: accumulated_msg - byte array, the message that was accumulated so far
# Return: 0 - if no more bytes are needed to complete the message
#         positive number - number of bytes that are needed to complete the message
#         negative number - invalid message detected
from common import iCOMOX_messages


def on_get_in_message_size(accumulated_msg, adxl356_smip=False):
    if len(accumulated_msg) == 0:  # if no byte was accumulated so far, then we need the IN_MSG_xxx.Code field
        return 1
    elif (accumulated_msg[0] > iCOMOX_messages.cCOMOX_MSG_CODE_COUNT) and (accumulated_msg[0] != iCOMOX_messages.cCOMOX_MSG_CODE_Report):
        return -1   # Invalid message code
    else:
        if accumulated_msg[0] < iCOMOX_messages.cCOMOX_MSG_CODE_COUNT:
            min_msg_size = iCOMOX_messages.COMOX_IN_MSG_SIZE[accumulated_msg[0]]
        else:
            min_msg_size = iCOMOX_messages.COMOX_IN_MSG_REPORT_HEADER_SIZE
        if len(accumulated_msg) < min_msg_size:
            return min_msg_size - len(accumulated_msg)
        elif accumulated_msg[0] != iCOMOX_messages.cCOMOX_MSG_CODE_Report:
            # Check if sCOMOX_IN_MSG_Debug.Cmd has a valid value
            if accumulated_msg[0] == iCOMOX_messages.cCOMOX_MSG_CODE_Debug:
                Cmd, Result = iCOMOX_messages.IN_MSG_Debug(msg=accumulated_msg)
                if Cmd > iCOMOX_messages.cCOMOX_DEBUG_NBIOT_CMD_TestConnectivity:
                    return -1
            return 0    # Here, for non report message we have done to receive a complete valid message
        else:
            module, sensor, axis, timestamp = iCOMOX_messages.IN_MSG_Report(msg=accumulated_msg)
            if module == iCOMOX_messages.cMODULE_RawData:
                if sensor >= iCOMOX_messages.cCOMOX_SENSOR_COUNT:
                    return -1   # Invalid IN_MSG_Report.payloadType
                if adxl356_smip:
                    if (sensor != iCOMOX_messages.cCOMOX_SENSOR_ADXL356) or (axis <= iCOMOX_messages.cAXIS_Z):  # if valid IN_MSG_Report.payloadType.sensor or axis fields
                        if (accumulated_msg[1] & 0x07) >= len(iCOMOX_messages.COMOX_IN_MSG_REPORT_PAYLOAD_SIZE_adxl356_smip):
                            return -1
                        return iCOMOX_messages.COMOX_IN_MSG_REPORT_HEADER_SIZE + iCOMOX_messages.COMOX_IN_MSG_REPORT_PAYLOAD_SIZE_adxl356_smip[accumulated_msg[1] & 0x07] - len(accumulated_msg)
                    else:
                        return -1  # invalid IN_MSG_Report.payloadType.sensor
                else:
                    return iCOMOX_messages.COMOX_IN_MSG_REPORT_HEADER_SIZE + iCOMOX_messages.COMOX_IN_MSG_REPORT_PAYLOAD_SIZE[accumulated_msg[1] & 0x07] - len(accumulated_msg)
            elif module == iCOMOX_messages.cMODULE_AnomalyDetection:
                return iCOMOX_messages.COMOX_IN_MSG_REPORT_HEADER_SIZE + 23 - len(accumulated_msg)
            # elif module == iCOMOX_messages.cMODULE_Maintenance:
            #     return iCOMOX_messages.COMOX_IN_MSG_REPORT_HEADER_SIZE + iCOMOX_messages.cCOMOX_SENSOR_COUNT*10 + 15 - len(accumulated_msg)
            elif module == iCOMOX_messages.cMODULE_Debug:
                return iCOMOX_messages.COMOX_IN_MSG_REPORT_HEADER_SIZE + iCOMOX_messages.DEBUG_REPOR_PAYLOAD_SIZE - len(accumulated_msg)
            else:
                return -1


def AXIS_to_Str(Axis):
    if Axis == iCOMOX_messages.cAXIS_X:
        return "X"
    elif Axis == iCOMOX_messages.cAXIS_Y:
        return "Y"
    elif Axis == iCOMOX_messages.cAXIS_Z:
        return "Z"
    else:
        return None

def iCOMOX_BoardType_to_Str(board_type):
    if board_type == iCOMOX_messages.cCOMOX_BOARD_TYPE_SMIP:
        return "SMIP"
    elif board_type == iCOMOX_messages.cCOMOX_BOARD_TYPE_NB_IOT:
        return "NB-IOT"
    elif board_type == iCOMOX_messages.cCOMOX_BOARD_TYPE_POE:
        return "POE"
    else:
        return "Unknown (0x{:02X})".format(board_type)

def iCOMOX_BoardVersion_to_Str(board_version_major, board_version_minor):
    if (board_version_major == 0) and (board_version_minor == 0):
        return ""
    else:
        return "{}.{}".format(board_version_major, board_version_minor)


def iCOMOX_Sensor_to_Str(sensor):
    if sensor < iCOMOX_messages.cCOMOX_SENSOR_COUNT:
        return iCOMOX_messages.COMOX_SENSOSRS_NAMES[sensor]
    else:
        return ""


def iCOMOX_firmware_release_version_branch_to_Str(firmware_release_version_branch):
    return ""
    # if firmware_release_version_branch == 0:
    #     return "DA Kit"
    # elif firmware_release_version_branch == 1:
    #     return "Suitcase"
    # else:
    #     return ""

def get_IN_MSG_Description(msg):
    if len(msg) < 1:
        return ""
    elif msg[0] == iCOMOX_messages.cCOMOX_MSG_CODE_Hello:
        return "Hello"
    elif msg[0] == iCOMOX_messages.cCOMOX_MSG_CODE_Reset:
        return "Reset"
    elif msg[0] == iCOMOX_messages.cCOMOX_MSG_CODE_GetConfiguration:
        return "GetConfiguration"
    elif msg[0] == iCOMOX_messages.cCOMOX_MSG_CODE_SetConfiguration:
        return "SetConfiguration"
    elif msg[0] == iCOMOX_messages.cCOMOX_MSG_CODE_Debug:
        return "Debug"
    elif msg[0] == iCOMOX_messages.cCOMOX_MSG_CODE_Report:
        if len(msg) < 2:
            return "Report"
        desc = "Report."
        module, sensor, axis, timestamp = iCOMOX_messages.IN_MSG_Report(msg)
        if module == iCOMOX_messages.cMODULE_RawData:
            if sensor == iCOMOX_messages.cCOMOX_SENSOR_ADXL362:
                desc += "LowPowerAcc"
            elif sensor == iCOMOX_messages.cCOMOX_SENSOR_ADXL356:
                desc += "Accelerometer."
                if axis <= iCOMOX_messages.cAXIS_Z:
                    desc += AXIS_to_Str(Axis=axis)
                else:
                    desc += "Invalid axis"
            elif sensor == iCOMOX_messages.cCOMOX_SENSOR_BMM150:
                desc += "Magnetometer"
            elif sensor == iCOMOX_messages.cCOMOX_SENSOR_ADT7410:
                desc += "Thermometer"
            elif sensor == iCOMOX_messages.cCOMOX_SENSOR_IM69D130:
                desc += "Microphone"
            else:
                desc += "Invalid"
            return desc
        elif module == iCOMOX_messages.cMODULE_AnomalyDetection:
            return "Anomaly Detection report"
        elif module == iCOMOX_messages.cMODULE_Maintenance:
            return "Maintenance report"
        elif module == iCOMOX_messages.cMODULE_Debug:
            return "Debug report"
        else:
            return "Unrecognized module"
    else:
        return "Unrecognized"


def eCOMOX_RESULT_to_Str(eCOMOX_RESULT):
    if eCOMOX_RESULT == iCOMOX_messages.cCOMOX_RESULT_OK:
        return "OK"
    elif eCOMOX_RESULT == iCOMOX_messages.cCOMOX_RESULT_SD_CARD:
        return "Error in saving file to SD card"
    else:
        return "Unknown error"

def eCOMOX_DEBUG_NBIOT_CMD_to_Str(eCOMOX_DEBUG_NBIOT_CMD):
    if eCOMOX_DEBUG_NBIOT_CMD == iCOMOX_messages.cCOMOX_DEBUG_NBIOT_CMD_Send_AT_Command:
        return "Send AT command"
    elif eCOMOX_DEBUG_NBIOT_CMD == iCOMOX_messages.cCOMOX_DEBUG_NBIOT_CMD_Reset_BG96:
        return "Reset modem"
    elif eCOMOX_DEBUG_NBIOT_CMD == iCOMOX_messages.cCOMOX_DEBUG_NBIOT_CMD_TestConnectivity:
        return "Test connectivity"
    else:
        return "Unknown command"

def ADC_Unpack_12bits_Array(packed_bytes):  # Unpack the 12 bits that are stored in 16 bits elements (4 MSB bits of each element are 0)
    unpacked_bytes = bytearray(len(packed_bytes) * 4 // 3)
    src_index = 0  # pointer to compress_samples
    dst_index = 0  # pointer to unpacked_bytes
    for i in range(0, len(unpacked_bytes)//4):
        unpacked_bytes[dst_index] = packed_bytes[src_index]
        unpacked_bytes[dst_index + 1] = packed_bytes[src_index + 1] & 0x0F
        unpacked_bytes[dst_index + 2] = (packed_bytes[src_index + 1] >> 4) + ((packed_bytes[src_index + 2] & 0x0F) << 4)
        unpacked_bytes[dst_index + 3] = packed_bytes[src_index + 2] >> 4
        dst_index += 4
        src_index += 3
    return unpacked_bytes

def Config_Options_Split(Options):
    if (Options & 0x01) != 0:
        CommChannel = iCOMOX_messages.cCOMOX_CONFIGURATION_COMM_CHANNEL_AUX
    else:
        CommChannel = iCOMOX_messages.cCOMOX_CONFIGURATION_COMM_CHANNEL_USB
    VibratorEnabled = (Options & 0x02) != 0
    TransmitEnabled = (Options & 0x40) != 0
    SaveToFileEnabled = (Options & 0x80) != 0
    return CommChannel, VibratorEnabled, TransmitEnabled, SaveToFileEnabled

def Config_Options_Build(CommChannel, VibratorEnabled, TransmitEnabled, SaveToFileEnabled):
    Options = 0
    if CommChannel == iCOMOX_messages.cCOMOX_CONFIGURATION_COMM_CHANNEL_AUX:
        Options |= 0x01
    if VibratorEnabled:
        Options |= 0x02
    if TransmitEnabled:
        Options |= 0x40
    if SaveToFileEnabled:
        Options |= 0x80
    return Options


"""
This module defines the helper function to draw the configuration
panel in the browser.
"""
# =====================imports=====================================
import streamlit as st
import os
import gateway_mocking as mocking
from txp.devices.ux.remote_gateway_client import remote_gateway_client
from txp.devices.gateway_states_enums import *
import txp.devices.ux.ux_states_protocol as ux_states
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==================== Functions ==========================================
def _print_details(details: ux_states.UxDetails):
    if details.kind == ux_states.UxDetails.Kind.Warning:
        st.warning(details.message)

    elif details.kind == ux_states.UxDetails.Kind.Info:
        st.info(details.message)
    else:
        st.error(details.message)


def _render_logical_configuration():
    """Renders the logical configuration where called."""
    st.subheader("Máquinas")
    for machine in st.session_state.gateway_state.machines:
        exp = st.expander(machine.machine_id)
        exp.markdown(f"**Modelo**: {machine.model}")
        exp.markdown(f"**Fabricante**: {machine.manufacturer}")
        exp.markdown(f"**Dispositivos configurados:**")
        for edge_id in machine.edges_ids:
            exp.write(edge_id)

    st.subheader("Dispositivos Edge")
    for edge in st.session_state.gateway_state.edges:
        exp = st.expander(edge.logical_id)
        exp.markdown(f"**Dispositivo**: {edge.device_type}")
        exp.markdown(f"**Tipo de Edge**: {edge.device_kind}")
        exp.markdown(f"**Parámetros de configuración**:")
        exp.write(edge.edge_parameters)


def user_confirmed_settings_callback():
    """Callback to be called when the user confirms to proceed from
    ConfigurationValidation state to DevicePairing state"""
    gateway_state = st.session_state.gateway_state
    sent_state = ux_states.UxConfigurationValidationState(
        gateway_state.state, ConfigurationValidationStatus.UserConfirmedSettings,
        None, gateway_state.gateway_conf, gateway_state.machines, gateway_state.edges
    )
    remote_gateway_client.send_state_to_gateway(
        sent_state
    )
    log.info('Sending UxConfigurationValidation-UserConfirmedSettings to Gateway.')


def render_configuration_panel():
    """The General Panel is used to show Gateway general configuration information,
    and to server the UX in the initial states of the Gateway when configuration
    is required.

    Specifically, the general information shown in this panel is:
        - Gateway metadata such as serial and project ID.
        - Logical configuration of the Gateway divided in sections for:
            Machines, Edges (After receiving the configuration from cloud).

    Specifically, the UX shown in this panel is:
        - User input of initial files and initial values.
        - Info about the logical configuration receiving process.
        - Info about the configuration validation process.
    """
    gateway_state = st.session_state.gateway_state
    st.markdown("-------------")
    st.subheader("Configuración del Gateway")
    if not hasattr(gateway_state, 'gateway_conf'):
        st.markdown(f"**Serial**: {gateway_state.gateway_id}")
        st.markdown(f"**Registro**: {gateway_state.registry_id}")
    else:
        st.markdown(f"**Serial**: {gateway_state.gateway_conf.gateway_id}")
        st.markdown(f"**Registro**: {gateway_state.gateway_conf.registry_id}")

    # Here starts the logic for the current state to follow in the Configuration Panel

    """
    Configuration Panel - Startup State. 
    In this state, the UX should render the required input values until all of them 
    have been received. Specifically:
        - First: The Private Key certificate file of the Gateway.
        - In any order: Text inputs for:
            - Gateway Serial 
            - IoT Registry 
    """
    if isinstance(gateway_state, ux_states.UxStartupState):
        st.markdown("### Configurar información inicial")

        """
        This placeholder is used to show markdown messages dynamically in this 
        position, later on. 
        """
        st.session_state.startup_markdown_placeholder = st.empty()

        if gateway_state.details:
            _print_details(gateway_state.details)
            gateway_state.details = None  # Delete details to continue interaction

        if not gateway_state.private_key_received:
            st.session_state.private_key_file_uploader_placeholder = st.empty()
            uploadedfile = (
                st.session_state.private_key_file_uploader_placeholder.file_uploader(
                    label="Subir el archivo de clave privada del Gateway",
                    type=".pem",
                    help="El Gateway necesita configurar su clave privada",
                )
            )
            if uploadedfile is not None:
                with open(
                    os.path.abspath(settings.gateway.private_key_path), "wb"
                ) as f:
                    f.write(uploadedfile.getbuffer())
                    f.close()
                gateway_state.private_key_received = True
                st.session_state.private_key_file_uploader_placeholder.empty()
                del st.session_state.private_key_file_uploader_placeholder

        if gateway_state.private_key_received:
            if not gateway_state.gateway_serial_received:
                st.session_state.gateway_serial_input_placeholder = st.empty()
                gateway_serial = (
                    st.session_state.gateway_serial_input_placeholder.text_input(
                        label="Introduzca el serial del Gateway",
                        key="gateway_serial_input_text",
                        value="",
                    )
                )
                if gateway_serial != "":
                    # Set Mocking value. TODO: Remove when integrating.
                    mocking.gateway_id = gateway_serial
                    gateway_state.gateway_id = gateway_serial
                    gateway_state.gateway_serial_received = True
                    st.session_state.gateway_serial_input_placeholder.empty()
                    del st.session_state.gateway_serial_input_placeholder

            if not gateway_state.registry_name_received:
                st.session_state.registry_id_input_placeholder = st.empty()
                registry_name = st.session_state.registry_id_input_placeholder.text_input(
                    label="Introduzca el nombre del registro",
                    key="registry_name_input_text",
                    value="",
                    help="El nombre del registro en la nube en el cual está registrado el Gateway",
                )
                if registry_name != "":
                    mocking.registry_id = registry_name
                    gateway_state.registry_id = registry_name
                    gateway_state.registry_name_received = True
                    st.session_state.registry_id_input_placeholder.empty()
                    del st.session_state.registry_id_input_placeholder

            if not gateway_state.project_id_received:
                st.session_state.project_id_input_placeholder = st.empty()
                project_id = st.session_state.project_id_input_placeholder.text_input(
                    label="Introduzca el nombre del proyecto GCP",
                    key="registry_name_input_text",
                    value="",
                    help="El nombre del proyecto en GCP en la nube en el cual está registrado el Gateway",
                )
                if project_id != "":
                    gateway_state.project_id = project_id
                    gateway_state.project_id_received = True
                    st.session_state.project_id_input_placeholder.empty()
                    del st.session_state.project_id_input_placeholder

        with st.session_state.startup_markdown_placeholder.container():
            if gateway_state.private_key_received:
                st.markdown("Clave privada configurada :white_check_mark:")

            if gateway_state.gateway_serial_received:
                st.markdown("Serial configurado :white_check_mark:")

            if gateway_state.registry_name_received:
                st.markdown("Registro en la nube configurado :white_check_mark:")

            if gateway_state.project_id_received:
                st.markdown("ID del proyecto configurado :white_check_mark:")

        # Show that the information is being sent to the Gateway.
        # This step is real, it's currently integrated with the Gateway process.
        if (
            gateway_state.private_key_received
            and gateway_state.gateway_serial_received
            and gateway_state.registry_name_received
            and gateway_state.project_id_received
        ):
            new_state = ux_states.UxStartupState(
                gateway_state.state,
                StartupStatus.RequiredValuesReceived,
                None,
                gateway_state.gateway_id,
                gateway_state.registry_id,
                gateway_state.private_key_received,
                gateway_state.gateway_serial_received,
                gateway_state.registry_name_received,
                gateway_state.project_id_received,
                gateway_state.project_id
            )
            remote_gateway_client.send_state_to_gateway(new_state)
            st.info("La información está siendo procesada")

    # In configuration Reception, the Gateway will be waiting for the cloud messages.
    # It will send updates to the UxState about that progress.
    elif isinstance(gateway_state, ux_states.UxConfigurationReceptionState):
        if not gateway_state.all_received:
            st.markdown("-------------")
            st.session_state.startup_markdown_placeholder = st.empty()
            st.info(
                f"Recibiendo configuraciones remotas."
            )
            with st.session_state.startup_markdown_placeholder.container():
                if (
                    gateway_state.status
                    == ConfigurationReceptionStatus.EstablishCloudConnection
                ):
                    st.markdown("Estableciendo conexión con la nube:exclamation:")
                else:
                    st.markdown("Conexión con la nube establecida:white_check_mark:")

                if gateway_state.gateway_conf.machines_ids:
                    st.markdown("Configuración del Gateway recibida:white_check_mark:")
                else:
                    st.markdown("Esperando por configuración lógica :exclamation:")

                if gateway_state.machines:
                    st.markdown(
                        "Configuración de Máquinas recibidas :white_check_mark:"
                    )
                else:
                    st.markdown("Esperando por configuración de Máquinas :exclamation:")

                if gateway_state.edges:
                    st.markdown(
                        "Configuración de dispositivos Edge recibida :white_check_mark:"
                    )
                else:
                    st.markdown(
                        "Esperando por configuración de dispositivos Edge :exclamation:"
                    )

    # In the ConfigurationValidation state, the Gateway will be processing the received
    # logical configurations, and checking if there were previous configurations
    # and if those configurations are outdated
    elif isinstance(gateway_state, ux_states.UxConfigurationValidationState):
        st.info(f"Configuraciones recibidas.")
        _render_logical_configuration()

        # Rendering the button with a container. Doing so, it's possible
        # to clean the button while the gateway sends the new state, and avoid
        # a problematic interaction
        button_container = st.empty()
        pressed = button_container.button(
            "Continuar",
            key="continue_to_pairing_button",
            on_click=user_confirmed_settings_callback,
        )
        if pressed:
            button_container.info("Preparando emparejamiento") # Clears the button and shows the message

    else:  # In this case, Configuration states were passed an only shows the configured values
        _render_logical_configuration()




"""
This module implements mcp9600 driver for a generic thermocouple.
"""
# ============================ imports =========================================

import threading
from typing import List
from txp.devices.drivers.driver import DriverState
from txp.common.edge import EdgeDescriptor

from txp.devices.package_queue import PackageQueueProxy
from txp.common.edge.perception_dimensions import TemperatureSignal
from txp.devices.drivers.driver import Driver
from txp.common.config import settings
import logging
import mcp9600
import queue

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# ==============================================================================
# Thermocouple Driver Implementation.
# ==============================================================================


class Mcp9600Driver(Driver):
    """Implementation of the mcp9600 driver for a generic thermocouple.
    """

    def __init__(
            self,
            edge_descriptor: EdgeDescriptor,
            gateway_packages_queue: PackageQueueProxy,
            on_driver_state_changed: callable,
            host_clock_notification_queue: queue.Queue
    ):
        """
        Note: This Driver receives a thermocouple setup, which is added and passed down by the Gateway operative
        configuration. The physical_id is not used by the driver because is assumed tha the mcp9600 is always connected
        to the I2C bus.

        Args
            edge_descriptor: The Edge Descriptor for this Driver.
            gateway_packages_queue: The Gateway Package Queue to put produced packages.
            on_driver_state_changed: Function used to notify driver state changes.
        """
        Driver.__init__(self, edge_descriptor, gateway_packages_queue, on_driver_state_changed,
                        host_clock_notification_queue)

        self.tt = None

    def observation_pulse(self) -> int:
        return 5

    @classmethod
    def device_name(cls) -> str:
        return "Generic"

    @classmethod
    def is_virtual(cls) -> bool:
        return False

    def connect_device(self):
        """Connects to the mcp9600.
        """
        try:
            self.tt = mcp9600.MCP9600()
            self.on_driver_state_changed(
                self.logical_id, DriverState.CONNECTED
            )
        except:
            self.on_driver_state_changed(
                self.logical_id, DriverState.DISCONNECTED
            )

        #self.tt.set_thermocouple_type('K')

    def disconnect_device(self):
        """Disconnects the mcp9600 and release the handler.
        """
        log.info(f"Thermocouple disconnected: {self.logical_id}")
        self.on_driver_state_changed(
            self.logical_id, DriverState.DISCONNECTED
        )

    def _start_sampling_collect(
            self
    ) -> List["Signals"]:
        """Collects the signals from the camera and generates the packages for the gateway.

        This method should collect samples that belong to the defined Sampling Time.

        Args:
            sampling_time_finished: event to check when the sampling time has finished.
        Returns:
            A list of ImageSignal collected by the mcp9600.
        """
        signals = []

        temperature = self.tt.get_hot_junction_temperature()

        if temperature is not None:
            signal = TemperatureSignal.build_signal_from_value(
                temperature,
                0
            )
            signals.append(signal)
        else:
            log.error('Error reading thermocouple connected mcp9600 ')

        return signals

    def _stop_sampling_collect(self):
        """This method will stop the sampling collection in the driver implementation.
        """
        pass



from txp.ml.common.tasks.vibration_basic_task import *
from txp.ml.common.tasks.slic_cv_tasks import SlicPatchRecognitionTask
from txp.ml.training_service.model_trainers.vibration_model_trainer import (
    RayVibrationModelTrainer,
)
from txp.ml.training_service.model_trainers.computer_vision_trainer import (
    RayComputerVisionSlicClassificationTrainer,
)
import logging
import json
from google.oauth2 import service_account
import google.cloud.firestore as firestore
from txp.common.ml.tasks import AssetTask


class TrainerFactory:
    # TODO: We should ensure that all the values in DB are coherent
    #   with the class names only. This will give us safety.
    _VIBRATION_MODELS = [
        VibrationBasicTask.__class__.__name__,
        "VibrationBasicGradientBoostTask",
        "VibrationBasicDecisionTreeTask",
    ]

    # TODO: We should ensure that all the values in DB are coherent
    #   with the class names only. This will give us safety.
    _COMPUTER_VISION_SLIC_MODELS = [
        SlicPatchRecognitionTask.__class__.__name__,
        'SlicPatchRecognitionTask'
    ]

    def __init__(self, credentials_str, models_bucket_name, data_warehouse_bucket_name, prediction_service_topic,
                 models_registry_collection_name, log_level=logging.INFO):

        logging.basicConfig(level=log_level)
        self._log_level = log_level

        self.models_bucket_name = models_bucket_name
        self.data_warehouse_bucket_name = data_warehouse_bucket_name
        self.prediction_service_topic = prediction_service_topic
        self.models_registry_collection_name = models_registry_collection_name
        self._credentials_str = credentials_str

        json_dict_service_account = json.loads(credentials_str, strict=False)
        credentials = service_account.Credentials.from_service_account_info(
            json_dict_service_account
        )
        self._credentials_str = credentials_str
        self._firestore_db = firestore.Client(
            credentials=credentials, project=credentials.project_id
        )

    def get_task_and_trainer(self, task_definition: AssetTask):
        logging.info(f"{self.__class__.__name__} finding task for {task_definition.dict()}.")
        task_type = task_definition.task_type
        if task_type in self._VIBRATION_MODELS:
            showroom_task = VibrationBasicTask(task_definition=task_definition, decision_tree_policy=False)
            vibration_trainer = RayVibrationModelTrainer.remote(
                self._credentials_str,
                self.models_bucket_name,
                self.data_warehouse_bucket_name,
                self.prediction_service_topic,
                self.models_registry_collection_name,
                self._log_level,
            )
            return showroom_task, vibration_trainer

        elif task_type in self._COMPUTER_VISION_SLIC_MODELS:
            logging.info(
                f"{self.__class__.__name__} received train request for " f"{task_type}."
            )
            slic_task = SlicPatchRecognitionTask(task_definition)
            trainer = RayComputerVisionSlicClassificationTrainer.remote(
                self._credentials_str,
                self.models_bucket_name,
                self.data_warehouse_bucket_name,
                self.prediction_service_topic,
                self.models_registry_collection_name,
                self._log_level,
            )

            return slic_task, trainer

        else:
            logging.error(f"Could not match any task for task_type: {task_type}")
            return None, None



import numpy as np
from common import messages_utils, iCOMOX_messages
from scipy.signal import welch
from scipy.signal.windows import hamming
import struct


class class_ADXL356:
    def __init__(self):
        # Initializing Fs_Acc, MFFT_Acc, window
        Root_clk = 26e6
        ACLKDIV = 8  # ADC clock is determined by Root_clk/ACLKDIV
        SAMPLETIME = 61 #24+10  # AcquisitionTime-1 in terms of ADC clocks
        OverSMP = 4  # Over sampling to increase resolution, or averaging which keeps resolution in digital domain
        NumOfCh = 3 #4  # Number of interleaved ChannelsBitmask
        DLY = 0 #65+63  # Delay at the end of chnnales round, in units of ADC clocks

        if DLY == 0:
            self.Fs_Acc = (Root_clk / ACLKDIV) / ((14 + SAMPLETIME) * OverSMP * NumOfCh)
        else:
            if NumOfCh == 1:
                self.Fs_Acc = (Root_clk / ACLKDIV) / (((14 + SAMPLETIME) * OverSMP) + DLY + 2)
            else:
                self.Fs_Acc = (Root_clk / ACLKDIV) / (((14 + SAMPLETIME) * OverSMP * NumOfCh) + DLY + 3)
        self.NFFT_Acc = 1024 #iCOMOX_messages.ADXL356_SAMPLES_NUM_PER_PIN    # 512
        self.Sensitivity_Acc = 80E-3  # 80mV/g for range of +/-10g. 40mV/g for range of +/-20g. 0g is translated to 900mV
        self.window = hamming(M=self.NFFT_Acc)

    def ADC_readings_to_g_units(self, ADC_samples):
        # convert ADC samples to g units
        return np.array(ADC_samples) * 1.8 / (4096 * self.Sensitivity_Acc)

    def g_units_to_PSD(self, acc_units_g):
        (freq_Acc, Pxxf_Acc) = welch(x=acc_units_g, fs=self.Fs_Acc, window=self.window, noverlap=self.NFFT_Acc/2, nfft=self.NFFT_Acc, return_onesided=True, detrend=False)
        return freq_Acc, Pxxf_Acc

    def msg_bytes_to_samples_g(self, payload):
        unpacked_bytes = messages_utils.ADC_Unpack_12bits_Array(packed_bytes=payload)
        ADC_samples = np.array(struct.unpack("<{}H".format(iCOMOX_messages.ADXL356_SAMPLES_NUM_PER_PIN*3), unpacked_bytes)) - 2048
        ADC_samples = self.ADC_readings_to_g_units(ADC_samples=ADC_samples)
        return ADC_samples[0::3], ADC_samples[1::3], ADC_samples[2::3]
        # axis_len = len(ADC_samples) // 3
        # return ADC_samples[0:axis_len], ADC_samples[axis_len:2*axis_len], ADC_samples[2*axis_len:]




from sensors import ADXL362
from sensors import ADXL356
from sensors import ADXL356_SMIP
from sensors import BMM150
from sensors import IM69D130
from sensors import ADT7410
from sensors import ADXL1002
#from communications import TCP_connectivity
from common import iCOMOX_list
from common import common_symbols

common_symbols.InitializeConfiguration()    # Read configuration constants from the INI file

app = None
#IPs = TCP_connectivity.GetIpAddresses()
iCOMOXs = iCOMOX_list.sCOMOXs_list()

ADXL362 = ADXL362.class_ADXL362()
ADXL356_SMIP = ADXL356_SMIP.class_ADXL356_SMIP()
ADXL356 = ADXL356.class_ADXL356()
BMM150 = BMM150.class_BMM150(quantile_for_noise_floor_estimator=common_symbols.BMM150_quantile_for_noise_floor_estimator, minimum_SNR_for_speed_detection_dB=common_symbols.BMM150_minimum_SNR_for_speed_detection_dB)
IM69D130 = IM69D130.class_IM69D130()
ADT7410 = ADT7410.class_ADT7410()
ADXL1002 = ADXL1002.class_ADXL1002()



"""
This module implements the driver host used to manage icomox POE devices through
a TCP/IP listener
"""
# ============================ imports =========================================
import multiprocessing
import queue
import threading
import time
from typing import Dict, List
from enum import Enum

from txp.devices.drivers.driver import DriverState, StateReport
from txp.devices.drivers.drivers_host import DriversHost
from txp.devices.drivers.icomox.icomox_datahandling import getHelloMessage, getIcomoxMessageType, IcomoxMessageType, \
    sendGetHelloMessageCommand, resetIcomox, sendSetConfigurationCommand
from txp.common.edge import EdgeDescriptor
from txp.devices.package_queue import PackageQueueProxy
from txp.devices.drivers.icomox.icomox_driver import IcomoxDriver
from txp.common.edge import DeviceId

from iCOMOXSDK.common.iCOMOX_list import sCOMOX_TCPIP
from iCOMOXSDK.communications.TCP_connectivity import class_iCOMOX_TcpServer

from txp.common.config import settings
from txp.common.local_ip import get_local_ip
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class TcpState(Enum):
    cTCP_STATE_DISCONNECTED = 0
    cTCP_STATE_LISTEN = 1
    cTCP_STATE_CLIENT_DISCONNECTED = 2
    cTCP_STATE_CLIENT_CONNECTED = 3
    cTCP_STATE_iCOMOX_CONNECTED = 4
    cTCP_STATE_ZOMBIE_CLIENT_DETECTED = 5


# ==============================================================================
# IcomoxDriversHost
# ==============================================================================
class IcomoxDriversHost(DriversHost):
    """Concrete implementation of a DriversHost for Icomox edge_descriptors.

    This class is executed as a separated process, and it's responsible for:
        - Start TCP/IP listener.
        - Host and configure icomox drivers connected to the listener
    """


    port = 1201
    connected_devices: Dict[str, str] = {}
    drivers_packets_queues: Dict[str, queue.PriorityQueue] = {}
    icomox_driver_errored_timeout_seconds = 15

    def __init__(self, gateway_packages_queue: PackageQueueProxy, drivers_state_queue: multiprocessing.Queue, **kwargs):

        """
        Args:
            gateway_packages_queue: The Gateway package queue used to collect the GatewayPackages.
            drivers_state_queue: The Gateway queue used to collect the driver states.
            **kwargs: named parameters specific to this IcomoxDriversHost implementation.
        """
        local_ip_add = get_local_ip()
        log.info(f"Icomox Hosts set ip_server_address to: {local_ip_add}")
        self.ip_server_address = local_ip_add

        log.info("Init TCP/IP listener for Icomox POE devices")
        super(IcomoxDriversHost, self).__init__(gateway_packages_queue,drivers_state_queue, self.icomox_driver_errored_timeout_seconds,**kwargs)

        try:
            self._tcpIpCommunication: class_iCOMOX_TcpServer = class_iCOMOX_TcpServer(
                callback_process_message = self._on_process_messages,
                callback_state_changed = self._on_tcpip_state_changed
                )

        except Exception as ex:
            log.error("Icomox SDK error creating class_iCOMOX_TcpServer")

    def connect(self):
        """Starts Icomox listener"""
        max_time_waiting_for_listener = 5

        self._tcpIpCommunication.start(self.ip_server_address, self.port)

        listener_time_finished = threading.Event()
        wait_for_server_timer = threading.Timer(
            max_time_waiting_for_listener,
            lambda e: e.set(),
            (listener_time_finished,),
        )

        wait_for_server_timer.start()

        while not self._tcpIpCommunication.is_listening() or listener_time_finished.isSet():
            time.sleep(0.5)

        if not self._tcpIpCommunication.is_listening():
            raise Exception("Error starting Icomox listener.")
        else:
            log.info("{} host is listening on address {} and port {}".format(self.get_driver_host_name(),self.ip_server_address,self.port))

    def disconnect(self):
        """Stops Icomox listener"""
        if self._tcpIpCommunication.is_listening():
            self._tcpIpCommunication.close()

    def discover_physical_ids(self) -> List[DeviceId]:
        """Returns the set of found DeviceId in the icomox network.

        Returns:
            List of device Ids
        """
        result = list()
        for value in self.connected_devices.values():
            if value != "":
                dev_id = DeviceId("", value)
                result.append(dev_id)
        return result

    def get_driver_host_name(self):
        return IcomoxDriver.device_name()

    def _add_drivers(self, edges: List[EdgeDescriptor]):
        """Adds the set of drivers for the specified edges in the host.

        Args:
            edges: List of EdgeDescriptor's for the drivers to be creted in
                the host instance.
        """
        #TODO: THIS CAN BE MOVED TO BASE CLASS. THE INSTANCE TYPE CAN BE CALCULLATED WITH NAME ATTRIBUTE OF THE HOST AND CREATOR CAN BE GOTTEN AS ATTRIBUTE AND ALL DRIVERS HANDLE SAME WAY
        for descriptor in edges:
            edge_queue = queue.PriorityQueue()
            self.drivers_packets_queues[descriptor.edge_parameters[IcomoxDriver._PHYSICAL_ID_PARAM_FIELD]] = edge_queue
            driver = IcomoxDriver(descriptor, self.gateway_package_queue, edge_queue, self._on_driver_state_changed,
                                  self._completed_driver_pulses_queue)
            self.drivers.append(driver)

    def take_fix_error_action(self, serial):
        for driver in self.drivers:
            if driver.logical_id == serial:
                resetIcomox(driver.get_complete_device_id().physical_id)

    def _on_process_messages(self, msg, iComox: sCOMOX_TCPIP =None):
        unique_id = self._get_unique_id_from_message(iComox, msg)
        self._add_message_to_queue(msg, unique_id)

    def _add_message_to_queue(self, msg, unique_id):
        # TODO: Check the correct timestamp
        timestamp = 1
        # Put the package in the driver priority queue, with the priority set to timestamp
        if self._sampling_started:
            if self.drivers_packets_queues.keys() is not None:
                if unique_id in self.drivers_packets_queues.keys():
                    self.drivers_packets_queues[unique_id].put((timestamp, msg))
                    log.debug("Package sent to driver")
                else:
                    log.warning('DEVICE NOT REGISTERED TO THIS GATEWAY {}'.format(unique_id))

    def _get_unique_id_from_message(self, iComox, msg):
        if iComox is None:
            raise Exception("on_process_messages(): iComox is None")
        unique_id = None
        if iComox.Hello is None:
            if getIcomoxMessageType(msg) == IcomoxMessageType.HELLO:
                iComox.Hello = msg
                hello_message = getHelloMessage(msg, iComox)
                # TODO: shall we add this parsed hello message to ICOMOX Driver Class as an attribute?
                if hello_message is not None:
                    unique_id = hello_message.getMcuSerialNumberString()
                    self._validate_icomox_physical_id_endpoints(unique_id)
                    self.connected_devices[iComox.remoteAddress] = unique_id
                    sendSetConfigurationCommand(unique_id,False) #making sure icomox will not send any data until it is requested to do it.
                    self._change_driver_state(unique_id, DriverState.CONNECTED)
            else:
                log.warning("ICOMOX device message is not Hello. Cant retrieve uniqueId")
                sendGetHelloMessageCommand(iComox)
        else:
            unique_id = iComox.UniqueIDString()
        return unique_id

    def _on_tcpip_state_changed(self, tcpState, iComox: sCOMOX_TCPIP = None):
        current_state = TcpState(tcpState)
        if current_state == TcpState.cTCP_STATE_CLIENT_CONNECTED:
            if iComox.remoteAddress not in self.connected_devices:
                self.connected_devices[iComox.remoteAddress] = "" #UniqueID not available yet
        elif current_state == TcpState.cTCP_STATE_CLIENT_DISCONNECTED:
            self._process_icomox_removal(iComox.remoteAddress)

    def _process_icomox_removal(self, address):
        if address in self.connected_devices:
            unique_id = self.connected_devices.pop(address)
            if unique_id != "":
                log.info("Icomox disconnected remote address {} and unique id {}".format(address, unique_id))
                self._change_driver_state(unique_id, DriverState.DISCONNECTED)
                self.clear_measured_enqueued_items(unique_id)

    def clear_measured_enqueued_items(self,physical_id):
        q = self.drivers_packets_queues[physical_id]
        try:
            while True:
                q.get_nowait()
                q.task_done()
        except queue.Empty:
            log.info("Queue for device {} physical ID has been cleared due to a disconnection".format(physical_id))
        except ValueError:
            log.info("Queue for device {} physical ID has been cleared due to a disconnection".format(physical_id))

    #TODO: we need to put down in base class of driver the physiical id artribute so we can generalize the behavior.
    def _change_driver_state(self, physical_id, driver_state):
        drivers = list(filter(lambda driver : driver.get_complete_device_id().physical_id == physical_id,self.drivers))
        if len(drivers) > 0:
            driver = drivers.pop()
            if driver_state  == DriverState.DISCONNECTED:
                driver.stop_sampling_windows()
            driver.update_driver_state(driver_state)
            self._add_drivers_state_to_queue(StateReport(driver.logical_id, driver_state))
        else:
            log.error("{} DEVICE IS NOT IN THE CURRENT  GATEWAY CONFIGURATION. ILLEGAL CONNECTION FROM SERIAL {}".format(self.get_driver_host_name(), physical_id))

    def _validate_icomox_physical_id_endpoints(self, unique_id):
        removable_address = None
        for ip,id in self.connected_devices.items():
            if id == unique_id:
                self._change_driver_state(unique_id,DriverState.DISCONNECTED)
                removable_address = ip
        if removable_address is not None:
            self._process_icomox_removal(removable_address)


"""
This module defines the StreamlitRunner class, which is a
custom class created to run the streamlit application programmatically
in another process.
"""
# ============================ imports =========================================
import streamlit.bootstrap as bootstrap
import txp.devices.ux.remote_gateway_client
from txp.common.config import settings as txp_settings
from txp.devices.ux.remote_gateway_client import remote_gateway_client
import multiprocessing as mp
import streamlit as st
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Definition of StreamlitRunner class
# ==============================================================================
class StreamlitRunner(mp.Process):
    """This class is the Process designed to start the Gateway UX
    streamlit application programmatically.

    It can receive the necessary objects to perform IPC communication
    with the main Gateway process.
    """

    def __init__(self, kwargs):
        """
        Args:
            kwargs: Mapping of str->values, expected to be used in the streamlit process
                app.
                gateway_connection: The multiprocess.Connection end of the pipe to be used
                    in the streamlit process.
        """
        super(StreamlitRunner, self).__init__(kwargs=kwargs)
        self.daemon = True
        self.name = "StreamlitRunnerProcess"

    def run(self) -> None:
        """Starts the Streamlit server application.

        It call the custom command call_streamlit_run, which initializes a CLI context
        and calls the main_run command from streamlit.

        It replicates the internal working of the _main_run method in streamlit.cli module:
            https://github.com/streamlit/streamlit/blob/d00429604bbc319bc44f3a326f8613607310978d/lib/streamlit/cli.py#L204

        This run method will keep running the streamlit server.
        """
        log.info("StreamlitRunner: Initialize streamlit process and objects")

        gateway_conn = self._kwargs["gateway_connection"]
        log.debug(f"StreamlitRunner: Received Gateway Connection: {id(gateway_conn)}")

        remote_gateway_client.set_gateway_connection(gateway_conn)

        remote_gateway_client.wait_for_initial_state()

        st._is_running_with_streamlit = True

        args = []
        flag_options = {}
        command_line = ""
        bootstrap.run(
            txp_settings.ux.streamlit_script, command_line, args, flag_options
        )



"""
This module declares the dataclass to be sent as the Gateway State to
the Cloud.
"""
# ============================ imports =========================================
from dataclasses import dataclass
from txp.devices.gateway_states_enums import GatewayStates
from txp.devices.drivers.driver import StateReport
from txp.common.json_complex_encoder import ComplexEncoderSerializable
from datetime import datetime
from typing import List, Dict
from txp.common.config import settings


@dataclass
class GatewayStatePayload(ComplexEncoderSerializable):
    """This GatewayStatePayload encapsulates the information sent to the cloud
    by the Gateway as the State report in the MQTT State topic.
    """
    state: GatewayStates
    edges_states: List[StateReport]
    boot_time: datetime
    ngrok_tcp_address: str
    local_ip: str
    tenant_id: str

    def reprJSON(self) -> Dict:
        edges_payload = {
            report.logical_id: {
                "driver_state": report.driver_state.human_readable_value,
                "since_datetime": report.since_datetime.strftime(
                    settings.gateway.tasks_scheduler.start_date_format
                )
            } for report in self.edges_states
        }

        return dict(
            tenant_id=self.tenant_id,
            edges=edges_payload,
            state=str(self.state),
            boot_time=self.boot_time,
            ngrok_tcp_address=self.ngrok_tcp_address,
            local_ip=self.local_ip
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'thirdparties'}

packages = \
['ArmSDK',
 'IpMgrConnectorMux',
 'IpMgrConnectorSerial',
 'SerialConnector',
 'SmartMeshSDK',
 'SmartMeshSDK.ApiDefinition',
 'SmartMeshSDK.IpMgrConnectorMux',
 'SmartMeshSDK.IpMgrConnectorSerial',
 'SmartMeshSDK.SerialConnector',
 'SmartMeshSDK.utils',
 'common',
 'iCOMOXSDK',
 'iCOMOXSDK.common',
 'iCOMOXSDK.communications',
 'iCOMOXSDK.sensors',
 'txp',
 'txp.cloud',
 'txp.cloud.cloud_functions',
 'txp.cloud.cloud_functions.paired_devices_cloud_update',
 'txp.cloud.cloud_functions.state_event_pubsub_to_firestore',
 'txp.cloud.cloud_functions.store_telemetry_event_to_data_lake',
 'txp.cloud.cloud_functions.store_telemetry_event_to_data_lake.test',
 'txp.cloud.cloud_functions.update_firestore_configuration',
 'txp.cloud.cloud_functions.update_iot_gateway_config',
 'txp.cloud.common',
 'txp.cloud.composer',
 'txp.cloud.controllers',
 'txp.cloud.pipelines',
 'txp.cloud.pipelines.batch_pipeline',
 'txp.cloud.pipelines.realtime_pipeline',
 'txp.cloud.pipelines.realtime_pipeline.test',
 'txp.cloud.schemas',
 'txp.cloud.scripts',
 'txp.cloud.scripts.firestore',
 'txp.cloud.scripts.identity_platform',
 'txp.cloud.states_events_anyscale_mocks',
 'txp.common',
 'txp.common.configuration',
 'txp.common.edge',
 'txp.common.ml',
 'txp.common.protos',
 'txp.devices',
 'txp.devices.drivers',
 'txp.devices.drivers.arm_robot',
 'txp.devices.drivers.cli_commands',
 'txp.devices.drivers.icomox',
 'txp.devices.drivers.mock',
 'txp.devices.drivers.thermocouple',
 'txp.devices.drivers.usb',
 'txp.devices.drivers.usb.camera',
 'txp.devices.drivers.voyager',
 'txp.devices.ux',
 'txp.ml',
 'txp.ml.common',
 'txp.ml.common.tasks',
 'txp.ml.common.tasks.slic_cv_tasks',
 'txp.ml.prediction_service',
 'txp.ml.prediction_service.schemas',
 'txp.ml.prediction_service.service_manager',
 'txp.ml.prediction_service.state_managers',
 'txp.ml.training_service',
 'txp.ml.training_service.dataset_fetchers',
 'txp.ml.training_service.model_trainers',
 'txp.web',
 'txp.web.core_components',
 'txp.web.resources',
 'txp.web.views',
 'txp.web.views.analytics']

package_data = \
{'': ['*'],
 'txp.cloud.cloud_functions.state_event_pubsub_to_firestore': ['dist/*'],
 'txp.cloud.cloud_functions.update_firestore_configuration': ['dist/*'],
 'txp.cloud.cloud_functions.update_iot_gateway_config': ['dist/*'],
 'txp.cloud.common': ['configs/*'],
 'txp.cloud.pipelines.realtime_pipeline': ['pub_sub_to_bigquery.egg-info/*'],
 'txp.common': ['configs/*'],
 'txp.devices.drivers.mock': ['resources/*'],
 'txp.ml': ['anyscale/*'],
 'txp.web.resources': ['images/*']}

install_requires = \
['Markdown>=3.3.6,<4.0.0',
 'PyJWT>=2.1.0,<3.0.0',
 'cryptography>=35.0.0,<36.0.0',
 'deepdiff>=5.7.0,<6.0.0',
 'dynaconf>=3.1.4,<4.0.0',
 'fastapi>=0.78.0,<0.79.0',
 'firebase-admin>=5.2.0,<6.0.0',
 'freezegun>=1.1.0,<2.0.0',
 'google-api-python-client>=2.23.0,<3.0.0',
 'google-auth-oauthlib>=0.4.6,<0.5.0',
 'google-cloud-firestore>=2.3.4,<3.0.0',
 'google>=3.0.0,<4.0.0',
 'ifaddr>=0.1.7,<0.2.0',
 'matplotlib',
 'mcp9600',
 'opencv-python-headless>=4.5.5,<5.0.0',
 'paho-mqtt>=1.5.1,<2.0.0',
 'plotly>=5.6.0,<6.0.0',
 'psutil>=5.8.0,<6.0.0',
 'pyserial>=3.5,<4.0',
 'pytz>=2022.1,<2023.0',
 'pyudev>=0.23.1,<0.24.0',
 'queuelib>=1.6.2,<2.0.0',
 'requests>=2.26.0,<3.0.0',
 'schedule>=1.1.0,<2.0.0',
 'scipy>=1.4.0,<2.0.0',
 'six==1.13.0',
 'sklearn>=0.0,<0.1',
 'streamlit==1.4.0',
 'tabulate>=0.8.9,<0.9.0',
 'transitions>=0.8.8,<0.9.0',
 'xgboost==1.6.0']

extras_require = \
{':python_full_version >= "3.7.1" and python_version < "4.0"': ['pandas==1.3.5',
                                                                'streamlit-aggrid>=0.2.3,<0.3.0'],
 ':python_version >= "3.7" and python_version < "3.11"': ['numpy>=1.21.2,<2.0.0',
                                                          'google-cloud-bigquery>=2.31.0,<3.0.0']}

entry_points = \
{'console_scripts': ['txp = txp.cli:txp_cli_entrypoint']}

setup_kwargs = {
    'name': 'txp',
    'version': '0.2.21',
    'description': 'Txp is the main Python Package from Tranxpert predictive maintenance solution',
    'long_description': '# TXP\n\n## Development\nThe following tools are used for this project:\n\n- [Poetry](https://python-poetry.org/) is used for dependency and package managment\n- [Nox](https://nox.thea.codes/en/stable/) is used as automation tool, mainly for testing\n- [Black](https://black.readthedocs.io/en/stable/) is the mandatory formatter tool\n- [PyEnv](https://github.com/pyenv/pyenv) is recommended as a tool to handle multiple python versions in developers machines. \n\n### Setup the development environment\n\n1. Install a supported Python version on your machine (compatible Python versions ^3.8). The recommended way is to use [PyEnv](https://github.com/pyenv/pyenv).\n\n2. Install the global Python required dependencies: \n\n    ```\n    pip install poetry nox\n    ```\n\n3. Clone this repository, and execute the following command in the repository root folder:\n\n    ```\n    poetry install\n    ``` \n    This will install all the dependencies for `txp` in a virtual enviroment created by Poetry for your project. \n    All the required dependencies for development are installed in that virtual enviroment. \n\n4. Configure your IDE to work with the virtual enviroment or the command line if you use an editor. \n\n    To activate the virtual enviroment on your terminal (MacOS, Linux) you can execute:\n    ```\n    source /path/to/poetry/cache/virtualenvs/test-O3eWbxRl-py3.7/bin/activate\n    ```\n\n    The path to your virtual enviroment location can be found with `poetry env info` [command](https://python-poetry.org/docs/managing-environments#displaying-the-environment-information). \n\n### Testing\n',
    'author': 'Tranxpert',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Traxpert/txp',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from txp.cloud import get_pubsub_event_log
from txp.ml.prediction_service.service_manager.factories import TaskFactory, StateManagerFactory
from txp.ml.prediction_service.service_manager.actors_synchronizer import stateful_function, ActorsStack, \
    stateful_method
import ray
import logging


@ray.remote
@stateful_function
def predict_and_save_event(task, state_manager, segregator, sampling_window, message, task_id, asset_id,
                           actors_stack=None, log_level=logging.INFO):
    """After TaskManager decided which ml task and state manager to use this ray remote function sends the current
    prediction to selected state manager.

    Args:
        task: corresponding task for incoming edge message.
        state_manager: corresponding state manager for asset in message.
        segregator: signals segregator object.
        sampling_window: All PubSub messages collected for given task.
        message: last PubSub message arrived.
        task_id: ml task_id.
        actors_stack: actors stack object.
        asset_id: asset or machine id
        log_level: log level for module
    """
    logging.basicConfig(level=log_level)
    sampling_window_rows = segregator.get_sampling_window_rows_from_bigquery(sampling_window,
                                                                             message["observation_timestamp"],
                                                                             message["tenant_id"])
    predicted_event = task.predict(sampling_window_rows)
    logging.info(f'Prediction completed for. {get_pubsub_event_log(message)}, prediction: {predicted_event}')
    event = {
        "event": predicted_event,
        "explainability": "{}",
        "task_id": task_id,
        "asset_id": asset_id,
        "observation_timestamp": message["observation_timestamp"],
        "event_id": f'{asset_id}-{task_id}-{message["observation_timestamp"]}',
        "tenant_id": message["tenant_id"],
        "partition_timestamp": message["partition_timestamp"]
    }

    logging.info(f"Processing event within state_manager: {get_pubsub_event_log(message)}")
    state_manager.process_event.remote(event, message["gateway_task_id"], message["sampling_window_index"],
                                       message["number_of_sampling_windows"], actors_stack=actors_stack)
    logging.info(f"Event processed {get_pubsub_event_log(message)}")


class TaskManager:
    """This class manages all ml tasks and state managers, basically this class decides which ml task and state
    manager to use by processing each PubSub message.

    Args:
        backup_collection_name: Firestore collection name where all backups are going to be stored.
        tenants_config: tenants configuration taken from firestore.
        events_and_states_dataset: Bigquery dataset where events and states are going to be stored.
        notifications_topic: PubSub topic id where notifications are going to be sent.
        _credentials_str: GCS credentials string.

    """

    def __init__(self, tenants_config, _credentials_str, backup_collection_name, events_and_states_dataset,
                 notifications_topic):
        self.actors_stack = ActorsStack.remote()
        self.tenants_config = tenants_config
        self.tasks = {}
        self.backup_collection_name = backup_collection_name
        self.events_and_states_dataset = events_and_states_dataset
        self.notifications_topic = notifications_topic
        self.state_managers = {}
        for tenant_id in self.tenants_config:
            self.state_managers = {**self.state_managers,
                                   **StateManagerFactory.create_state_managers(
                                       self.tenants_config[tenant_id], _credentials_str, self.backup_collection_name,
                                       self.events_and_states_dataset, self.notifications_topic)}
            self.tasks = {**self.tasks, **TaskFactory.create_tasks(self.tenants_config[tenant_id], self.state_managers,
                                                                   _credentials_str)}

    def backup(self):
        logging.info("Backing up task manager, all state managers present at the service will be backed up")
        for sm_id in self.state_managers:
            ray.get(self.state_managers[sm_id].backup.remote())

    @stateful_method
    def process_messages(self, messages, signals_segregator):
        current_task = None
        current_state_manager = None

        for message in messages:

            logging.info(f"Event PubSub received. {get_pubsub_event_log(message)}")

            task_id = None
            asset_id = None
            for tenant_id in self.tenants_config:
                for asset in self.tenants_config[tenant_id]:
                    for task in self.tenants_config[tenant_id][asset]['tasks'].values():
                        for edge in task['edges']:
                            if edge == message["edge_logical_id"]:
                                if self.tenants_config[tenant_id][asset]["state_manager"] not in self.state_managers \
                                        or task['task_id'] not in self.tasks:
                                    continue
                                asset_id = asset
                                task_id = task["task_id"]
                                current_task = self.tasks[task['task_id']]
                                current_state_manager = self.state_managers[
                                    self.tenants_config[tenant_id][asset_id]["state_manager"]]
            if current_task is None or current_state_manager is None:
                logging.info(
                    f"Current configuration is wrong not state manager or current task found for PubSub event. "
                    f"{get_pubsub_event_log(message)}")
                continue
            schema = current_task.get_schema()
            completed = signals_segregator.process_message(message, schema, task_id)

            if completed:
                logging.info(f"All necessary PubSub event were received when {get_pubsub_event_log(message)} "
                             f"was found, generating prediction.")

                sampling_window = signals_segregator.get_sampling_window(message["observation_timestamp"], task_id)
                signals_segregator.delete_sampling_window(message["observation_timestamp"], task_id)
                predict_and_save_event.remote(current_task, current_state_manager,
                                              signals_segregator.bigquery_segregator, sampling_window,
                                              message, task_id, asset_id, actors_stack=self.actors_stack)



from dataclasses import dataclass
from abc import ABC, abstractmethod


class BasicTrainingServiceException(ABC):

    @abstractmethod
    def get_message(self, **key_args):
        pass


@dataclass
class NoParquetFileFound(FileNotFoundError, BasicTrainingServiceException):
    file_name: str

    def get_message(self, **key_args):
        return f" Could not find parquet file {self.file_name} at {key_args['gcs_file_path']}"


@dataclass
class ErrorWhileTrainingModel(Exception):  # TODO: this is a too broad exception specific errors should be captured

    def get_message(self, **key_args):
        return f"There was an error during model training, please refer to service logs for more info"


@dataclass
class WrongLabelValue(Exception):
    label_value: str

    def get_message(self, **key_args):
        return f"Wrong label value processing data: {self.label_value}"




"""
    This module provides functions and classes that helps with programmatic interaction with
    IoT core service.
"""

from google.cloud import iot_v1
import logging
from datetime import datetime
from txp.common.config import settings


class IoTDevice:
    """A wrapper dataclass for the Device type from iot_v1 library:

    https://googleapis.dev/python/cloudiot/latest/iot_v1/types.html#google.cloud.iot_v1.types.Device
    """
    def __init__(self, device_response):
        self.heartbeat: datetime = datetime(
            device_response.last_heartbeat_time.year,
            device_response.last_heartbeat_time.month,
            device_response.last_heartbeat_time.day,
            device_response.last_heartbeat_time.hour,
            device_response.last_heartbeat_time.minute,
            device_response.last_heartbeat_time.second
        )

        self.last_config_ack: datetime = datetime(
            device_response.last_config_ack_time.year,
            device_response.last_config_ack_time.month,
            device_response.last_config_ack_time.day,
            device_response.last_config_ack_time.hour,
            device_response.last_config_ack_time.minute,
            device_response.last_config_ack_time.second
        )

        self.last_error_msg: str = device_response.last_error_status.message

        self.last_error_time: datetime = datetime(
            device_response.last_error_time.year,
            device_response.last_error_time.month,
            device_response.last_error_time.day,
            device_response.last_error_time.hour,
            device_response.last_error_time.minute,
            device_response.last_error_time.second
        )

        self.last_state_time: datetime = datetime(
            device_response.last_state_time.year,
            device_response.last_state_time.month,
            device_response.last_state_time.day,
            device_response.last_state_time.hour,
            device_response.last_state_time.minute,
            device_response.last_state_time.second
        )






log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


def get_authenticate_client(service_account_credentials) -> iot_v1.DeviceManagerClient:
    """Returns a Device Manager client authenticated to perform operations against Core IoT.

    The google.cloud.iot_v1.types.Device instance for the selected device.

    https://googleapis.dev/python/cloudiot/latest/iot_v1/device_manager.html
    """
    # Create a client
    client = iot_v1.DeviceManagerClient(
        credentials=service_account_credentials
    )
    return client


def get_device_state(
    iot_client: iot_v1.DeviceManagerClient,
    project_id: str,
    cloud_region: str,
    registry_id: str,
    gateway_id: str,
) -> IoTDevice:
    """
    Args:
        iot_client: The IoT core device client.
        project_id: The project ID to which the device belong.
        cloud_region:
        registry_id:
        gateway_id:

    Returns:

    """
    device_path = iot_client.device_path(
        project_id,
        cloud_region,
        registry_id,
        gateway_id
    )

    response = iot_client.get_device(
        name=device_path
    )

    if response:
        device_instance = IoTDevice(response)
        return device_instance
    else:
        return None










"""
This module contains the concrete implementation case of Driver for the
Voyager 3 device.
Technical details about the device can be found in the following document:
https://tranxpert.atlassian.net/wiki/spaces/TD/pages/230817807/Tranxpert+MVP+Monitoring+Proposal#Several-Voyager-3-working-together.
"""

# ============================ imports =========================================
import queue
import threading
import time
from typing import Dict, Tuple, List

from txp.devices.drivers.voyager.smartmesh_dongle import SmartMeshDongle
from txp.devices.drivers.driver import Driver, EdgeParameterNotFoundError
from txp.common.edge import EdgeDescriptor
from txp.devices.package_queue import PackageQueueProxy
from txp.common.edge import VibrationAccelerationSignal

from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# ==============================================================================
# Voyager driver implementation
# ==============================================================================
class VoyagerDriver(Driver):
    """Implementation of the Voyager 3 vibration sensor driver.

    The Voyager Driver is a logical unit configured in The IoT management.
    Operational physical units are mapped to the logical configuration by using
    a Logical ID -> Mac Address mapping. This allows the system to rotate and re assign
    physical sensors without problems.
    """

    def __init__(
            self,
            edge_descriptor: EdgeDescriptor,
            gateway_packages_queue: PackageQueueProxy,
            smartmesh_manager: SmartMeshDongle,
            on_driver_state_changed: callable,
            host_clock_notification_queue: queue.Queue
    ):
        """
        Note: This Driver receives a MAC address, which is added and passed down by the Gateway operative
        configuration. The MAC address is not known by the Cloud configuration.

        Args
            edge_descriptor: The Edge Descriptor for this Driver.
            gateway_packages_queue: The Gateway Package Queue to put produced packages.
            smartmesh_manager: The SmartMesh manager that interact with the physical device.
            on_driver_state_changed: Function used to notify driver state changes.
        """
        Driver.__init__(self, edge_descriptor, gateway_packages_queue,
                        on_driver_state_changed, on_driver_state_changed, host_clock_notification_queue)
        # The Voyager receives 3 perception dimensions with adc_num_samples length

        if self.PERCEPTION_DIMENSIONS_FIELD not in self.perceptions[VibrationAccelerationSignal.perception_name()]:
            log.error(
                f"VoyagerDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received in "
                f"{VibrationAccelerationSignal.perception_name()} perception configuration"
            )
            raise EdgeParameterNotFoundError(
                f"VoyagerDriver: {self.PERCEPTION_DIMENSIONS_FIELD} not received in "
                f"{VibrationAccelerationSignal.perception_name()} perception configuration"
            )

        if len(self.perceptions[VibrationAccelerationSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD]) != 2:
            log.error(
                f"VoyagerDriver: {self.PERCEPTION_DIMENSIONS_FIELD} in {VibrationAccelerationSignal.perception_name()} "
                f"perception should have length 2"
            )
            raise EdgeParameterNotFoundError(
                f"VoyagerDriver: {self.PERCEPTION_DIMENSIONS_FIELD} in {VibrationAccelerationSignal.perception_name()} "
                f"perception should have length 2"
            )

        if self.PERCEPTION_SAMPLING_FREQUENCY_FIELD not in self.perceptions[VibrationAccelerationSignal.perception_name()]:
            log.error(
                f"VoyagerDriver: {self.PERCEPTION_SAMPLING_FREQUENCY_FIELD} not received "
                f"{VibrationAccelerationSignal.perception_name()} perception configuration"
            )
            raise EdgeParameterNotFoundError(
                f"VoyagerDriver: {self.PERCEPTION_SAMPLING_FREQUENCY_FIELD} not received "
                f"{VibrationAccelerationSignal.perception_name()} perception configuration"
            )

        self.sampling_frequency = int(self.perceptions[VibrationAccelerationSignal.perception_name()][self.PERCEPTION_SAMPLING_FREQUENCY_FIELD])

        self._signal_shape: Tuple[int, int] = (
            int(self.perceptions[VibrationAccelerationSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][0]),
            int(self.perceptions[VibrationAccelerationSignal.perception_name()][self.PERCEPTION_DIMENSIONS_FIELD][1]),
        )
        """
        _signal_shape: Internal attribute to control the VibrationSignal dimensions.
        """

        self._manager: SmartMeshDongle = smartmesh_manager
        self.mac: Tuple = self._parse_mac_string(self._physical_id)
        self._mote_packets_queue: queue.PriorityQueue = queue.PriorityQueue()
        """
        _samples_queue: A queue provided to the Voyager SmartMesh manager to put 
            received notification packets, with the timestamp as priority.

            Since the Manager runs in the same DriversHost process as the Driver instance, 
            a threa-safe PriorityQueue can be shared.
        """

        self.setup_mote()

        self._lock = threading.Lock()

        # Internal control attributes
        self._data_valid = False
        self._rx_cnt = 0
        self._x_offset = 0
        self._y_offset = 0
        self._z_offset = 0
        self._offset_removed = {"x": False, "y": False, "z": False}

        # data buffer for num_samples for raw samples and num_samples/2 for fft magnitude
        self._data1 = int(
            (3 * self._manager.adc_num_samples / 2) + self._manager.adc_param_len
        ) * [0]
        self._data2 = int(
            (3 * self._manager.adc_num_samples / 2) + self._manager.adc_param_len
        ) * [0]

        # Ping-pong buffer, _data_order[0] = currently filling
        self._data_order = (self._data1, self._data2)
        self.scale = [0.75, 0.75, 0.75]

        self.remove_offsets = False
        self.offset = 0  # Default offset for all axis


    @staticmethod
    def _parse_mac_string(mac_str: str) -> Tuple:
        """Parse the expected mac string in hex to Tuple.

        Example:
            "00-17-0d-00-00-71-c6-48" to (0, 23, 13, 0, 0, 113, 198, 72)

        Returns:
            The tuple for the MAC address.
        """
        mac_string = mac_str.split("-")
        mac = tuple(int(i, 16) for i in mac_string)

        return mac

    @staticmethod
    def _pretty_mac(mac) -> str:
        """Parses mote mac tuple into is corresponding string

        Args:
            mac: mac id tuple
        Returns:
            Mac string.
        """
        pMac = "-".join(["%02x" % b for b in mac])
        return pMac

    def observation_pulse(self) -> int:
        60

    @classmethod
    def device_name(cls) -> str:
        return "Voyager"

    @classmethod
    def is_virtual(cls) -> bool:
        return False

    def setup_mote(self):
        frequency_msg = str(self.sampling_frequency).ljust(8, 'x')
        alarm_msg = "0xxxxxxx"
        axis_sel_msg = "111xxxxx"
        message = frequency_msg + alarm_msg + axis_sel_msg

        self._manager.send_to_mote(self.mac, message)

    def get_latest_valid_measure(self, with_offset=False):
        """Uses threading lock to return current frame of raw, fft and
        axis data.

        Args:
            with_offset(bool): If True, returns the raw and fft data without the
            offset removed.
        Returns:
            A tuple with the raw, fft and axis information from the last completed measure
            received by the smarthmesh manager.
        """
        if not self._data_valid:  # _handle_bytes has not yet been called
            return (None, None, None)

        with self._lock:
            params = self._data_order[1][0]  # Info on axis - x,y,z
            raw = self._data_order[1][2: self._manager.adc_num_samples]
            fft = self._data_order[1][
                  self._manager.adc_num_samples + self._manager.adc_param_len:
                  ]

        x_set = (int(params & 0x00F0)) == 144
        y_set = (int(params & 0x00F0)) == 160
        z_set = (int(params & 0x00F0)) == 192

        if x_set:
            ax = "x"
            self.offset = self._x_offset
        elif y_set:
            ax = "y"
            self.offset = self._y_offset
        elif z_set:
            ax = "z"
            self.offset = self._z_offset

        raw_avg = sum(raw) / len(raw)

        if with_offset:
            # raw = [ADC_gain * (x - ADC_offset) for x in raw]
            pass  # Just return raw codes
        else:
            raw = [
                (self._manager.ADC_gain * (x - self._manager.ADC_offset - self.offset))
                for x in raw
            ]

        raw_avg = sum(raw) / len(raw)
        # print("{} axis: Average G Value = {}".format(ax, raw_avg))

        fft = [2 * self._manager.ADC_gain * x for x in fft]

        return raw, fft, ax

    def connect_device(self):
        """The manager connects devices on its own. This 'connection' registers
        the driver _mote_packets_queue in the Manager.
        """
        self._device_connected()  # State transition.

    def disconnect_device(self):
        """Nothing to disconnect. The manager takes care of dealing with
        connected and unconnected devices.
        """
        self._manager.unregister_device(self.mac)
        self._device_disconnected()

    def _start_sampling_collect(
            self
    ) -> List["Signals"]:
        """This method will collect the samples obtained from the edge device and will
        produce the system Signals to send the package to the gateway.

        This method should collect samples that belong to the defined Sampling Time.

        Args:
            sampling_time_finished: event to check when the sampling time has finished.
        Returns:
            A list of VibrationSignal collected by the Voyager.
        """
        signals = []
        measures: Dict[str, list] = {}

        self._manager.register_driver(self.mac, self._mote_packets_queue)

        while True:
            if self._handle_packet():
                raw, fft, ax = self.get_latest_valid_measure()
                measures[ax] = raw

            if len(measures.keys()) == self._signal_shape[0]:
                signal = VibrationAccelerationSignal.build_signal_from_lists(
                    [measures.get('x'), measures.get('y'), measures.get('z')],
                    self._signal_shape,
                    self.sampling_frequency,
                    self._manager.adc_num_samples,
                )

                measures.clear()
                signals.append(signal)

                if len(signals) == 1:
                    break

        return signals

    def _stop_sampling_collect(self):
        """This method will stop the sampling collection in the driver implementation.
        """
        self._manager.unregister_device(self.mac)

    def _handle_packet(self) -> bool:
        """Handles the smartmesh data packages received by the smartmesh manager.
        The Voyager 3, splits one axis measurement in 17 smartmesh packages.

        Returns:
            True if a measurement was completed; otherwise False.
        """
        signal_ready = False
        timeout_in_sec = 2

        try:
            signal_ready = self._handle_bytes(self._mote_packets_queue.get(True, timeout_in_sec)[1])
        except queue.Empty:
            pass

        return signal_ready

    def _handle_bytes(self, bytes) -> bool:
        """Alings incoming SmartMesh bytes to form frames for analysis.
        Raw and fft variables are filled when method completes.

        Args:
            bytes: 90 byte packet passed down from handle_packet.
        Returns:
            True if a measurement was completed; otherwise False.
        """

        try:
            cur_data = self._data_order[0]
            data_len = len(bytes) / 2
            msb_set = bytes[1] == 255  # MSB of first word is alignment bit

            if self._rx_cnt + data_len > len(
                    cur_data
            ):  # Restart if we get more data in a frame than expected
                log.debug(
                    f"[{self.logical_id} {self._rx_cnt}] Restarting after data overflow"
                )
                self._rx_cnt = 0

            if (
                    self._rx_cnt > 0 and msb_set
            ):  # Restart if we get an alignment bit in the middle of a frame
                log.debug(f"[{self.logical_id}] Restarting after partial frame")
                log.debug(f"Rx_cnt {self._rx_cnt}, cur_data {len(cur_data)}")
                self._rx_cnt = 0

            if self._rx_cnt == 0 and not msb_set:  # Align to the start of a new frame
                log.debug(f"[{self.logical_id}] Waiting for start of new frame")
                return  # Continue count until full frame is reached

            self._lock.acquire()

            for i in range(0, len(bytes), 2):
                cur_data[self._rx_cnt] = (bytes[i + 1] << 8) | bytes[
                    i
                ]  # Convert from bytes to words
                self._rx_cnt = self._rx_cnt + 1

            frame_got = False  # True if we have entered the next if statement, avoids double release of lock
            # Are we done with frame?
            if self._rx_cnt == len(cur_data):
                self._data_order = (self._data_order[1], self._data_order[0])
                self._data_valid = True
                self._rx_cnt = 0
                self._lock.release()
                frame_got = True

                raw, fft, ax = self.get_latest_valid_measure()
                log.debug(
                    f"Got complete frame @ {time.process_time()} -- Axis: {ax} -- Mote ID: {self.logical_id}"
                )

                if self.remove_offsets:
                    self._remove_offset()
                    if self._offset_removed.values() == [True, True, True]:
                        self.remove_offsets = False

            if not frame_got:  # Ensure threading lock not released twice
                self._lock.release()
        except Exception as e:
            logging.error(e, exc_info=True)

        return frame_got

    def _remove_offset(self):
        """Updates mote offset variables for each axis, x, y, z by calculating mean for each.
        This function must be called just one time and the offset calculated will be applied
        to the measure on get_latest_valid_measure(with_offset=True).
        Recommended by the provider as calibration mechanims.
        """
        if not self._data_valid:  # _handle_bytes has not yet been called
            return

        raw_w_offset, fft, ax = self.get_latest_valid_measure(with_offset=True)

        if ax == "x" and self._offset_removed[ax] == False:
            self._x_avg0G = sum(raw_w_offset) / len(raw_w_offset)
            self._x_offset = self._x_avg0G - self._manager.ADC_offset
            self._offset_removed[ax] = True
            log.debug("Removing {} from X axis codes".format(self._x_offset))

        if ax == "y" and self._offset_removed[ax] == False:
            self._y_avg0G = sum(raw_w_offset) / len(raw_w_offset)
            self._y_offset = self._y_avg0G - self._manager.ADC_offset
            self._offset_removed[ax] = True
            log.debug("Removing {} from Y axis codes".format(self._y_offset))

        if ax == "z" and self._offset_removed["x"] and self._offset_removed["y"]:
            # Make sure x and y have been offset before Z
            self._z_offset = (
                                     self._x_avg0G + self._y_avg0G
                             ) / 2 - self._manager.ADC_offset
            self._offset_removed[ax] = True
            log.debug("Removing {} from Z axis codes".format(self._z_offset))
